#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2020 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
from math import sqrt

from adcc import block as b
from adcc.LazyMp import LazyMp
from adcc.AdcMethod import IsrMethod
from adcc.functions import einsum
from adcc.Intermediates import Intermediates
from adcc.AmplitudeVector import AmplitudeVector
from adcc.OneParticleDensity import OneParticleDensity
from adcc.NParticleOperator import OperatorSymmetry

from .util import check_doubles_amplitudes, check_singles_amplitudes


def diffdm_isr0(mp, amplitude, intermediates):
    # C is either c(ore) or o(ccupied)
    C = b.c if mp.has_core_occupied_space else b.o
    check_singles_amplitudes([C, b.v], amplitude)
    u1 = amplitude.ph

    dm = OneParticleDensity(mp, symmetry=OperatorSymmetry.HERMITIAN)
    dm[C + C] = -einsum("ia,ja->ij", u1, u1)
    dm.vv = einsum("ia,ib->ab", u1, u1)
    return dm


def diffdm_isr1(mp, amplitude, intermediates):
    dm = diffdm_isr0(mp, amplitude, intermediates)  # Get ISR(0) result

    try:
        # ISR(1)-d
        check_doubles_amplitudes([b.o, b.o, b.v, b.v], amplitude)
        u1, u2 = amplitude.ph, amplitude.pphh
        dm.ov += -2 * einsum("jb,ijab->ia", u1, u2)
    except ValueError:
        # no doubles contribution
        pass
    return dm


def diffdm_isr2(mp, amplitude, intermediates):
    dm = diffdm_isr1(mp, amplitude, intermediates)  # Get ISR(1) result
    check_doubles_amplitudes([b.o, b.o, b.v, b.v], amplitude)
    u1, u2 = amplitude.ph, amplitude.pphh

    t2 = mp.t2(b.oovv)
    p0 = mp.mp2_diffdm
    p1_oo = dm.oo.evaluate()  # ISR(1) diffdm
    p1_vv = dm.vv.evaluate()  # ISR(1) diffdm

    # Zeroth order doubles contributions
    p2_oo = -einsum("ikab,jkab->ij", u2, u2)
    p2_vv = einsum("ijac,ijbc->ab", u2, u2)
    p2_ov = -2 * einsum("jb,ijab->ia", u1, u2).evaluate()

    # ISR(2) intermediate (TODO Move to intermediates)
    ru1 = einsum("ijab,jb->ia", t2, u1).evaluate()

    # Compute second-order contributions to the density matrix
    dm.oo += (  # adc2_p_oo
        2 * p2_oo - einsum("ia,ja->ij", ru1, ru1) + (
            + einsum("ik,kj->ij", p1_oo, p0.oo)
            - einsum("ikcd,jkcd->ij", t2,
                     + 0.5 * einsum("lk,jlcd->jkcd", p1_oo, t2)
                     - einsum("jkcb,db->jkcd", t2, p1_vv))
            - einsum("ia,jkac,kc->ij", u1, t2, ru1)
        ).symmetrise()
    )

    dm.vv += (  # adc2_p_vv
        2 * p2_vv + einsum("ia,ib->ab", ru1, ru1) - (
            + einsum("ac,cb->ab", p1_vv, p0.vv)
            + einsum("klbc,klac->ab", t2,
                     + 0.5 * einsum("klad,cd->klac", t2, p1_vv)
                     - einsum("jk,jlac->klac", p1_oo, t2))
            - einsum("ikac,kc,ib->ab", t2, ru1, u1)
        ).symmetrise()
    )

    dm.ov += (  # adc2_p_ov
        - einsum("ijab,jb->ia", t2, p2_ov)
        - einsum("ib,ba->ia", p0.ov, p1_vv)
        + einsum("ij,ja->ia", p1_oo, p0.ov)
        - einsum("ib,klca,klcb->ia", u1, t2, u2)
        - einsum("ikcd,jkcd,ja->ia", t2, u2, u1)
    )
    return dm


def diffdm_cvs_isr1(mp, amplitude, intermediates):
    dm = diffdm_isr0(mp, amplitude, intermediates)  # Get ISR(0) result

    try:
        # ISR(1)-d
        check_doubles_amplitudes([b.o, b.c, b.v, b.v], amplitude)
        u1, u2 = amplitude.ph, amplitude.pphh
        p2_ov = -sqrt(2) * einsum("jb,ijab->ia", u1, u2)
        dm.ov += p2_ov
    except ValueError:
        # no doubles contribution
        pass
    return dm


def diffdm_cvs_isr2(mp, amplitude, intermediates):
    dm = diffdm_cvs_isr1(mp, amplitude, intermediates)  # Get cvs-ISR(1) result
    check_doubles_amplitudes([b.o, b.c, b.v, b.v], amplitude)
    u1, u2 = amplitude.ph, amplitude.pphh

    t2 = mp.t2(b.oovv)
    p0 = intermediates.cvs_p0
    p1_vv = dm.vv.evaluate()  # ISR(1) diffdm

    # Zeroth order doubles contributions
    p2_vo = -sqrt(2) * einsum("ijab,jb->ai", u2, u1)
    p2_oo = -einsum("ljab,kjab->kl", u2, u2)
    p2_vv = 2 * einsum("ijac,ijbc->ab", u2, u2)

    # Second order contributions
    # cvs_isr2_dp_oo
    dm.oo += p2_oo + einsum("ab,ikac,jkbc->ij", p1_vv, t2, t2)

    dm.ov += (  # cvs_isr2_dp_ov
        - einsum("ka,ab->kb", p0.ov, p1_vv)
        - einsum("lkdb,dl->kb", t2, p2_vo)
        + 1 / sqrt(2) * einsum("ib,klad,liad->kb", u1, t2, u2)
    )

    dm.vv += p2_vv - 0.5 * (  # cvs_isr2_dp_vv
        + einsum("cb,ac->ab", p1_vv, p0.vv)
        + einsum("cb,ac->ab", p0.vv, p1_vv)
        + einsum("ijbc,ijad,cd->ab", t2, t2, p1_vv)
    )

    # Add 2nd order correction to CVS-ISR(1) diffdm
    dm.cc -= einsum("kIab,kJab->IJ", u2, u2)
    return dm


# dict controlling the dispatch of the state_diffdm function
DISPATCH = {
    "isr0": diffdm_isr0,
    "isr1s": diffdm_isr0,   # Identical to ISR(0)
    "isr1": diffdm_isr1,
    "isr2d": diffdm_isr2,  # Identical to ISR(2)
    "isr2": diffdm_isr2,
    "cvs-isr0": diffdm_isr0,
    "cvs-isr1s": diffdm_isr0,   # Identical to ISR(0)
    "cvs-isr1": diffdm_cvs_isr1,
    "cvs-isr2d": diffdm_cvs_isr2,  # Identical to CVS-ISR(2)
    "cvs-isr2": diffdm_cvs_isr2,
}


def state_diffdm(method, ground_state, amplitude, intermediates=None):
    """
    Compute the one-particle difference density matrix of an excited state
    in the MO basis.

    Parameters
    ----------
    method : str, IsrMethod
        The method to use for the computation (e.g. "isr2")
    ground_state : LazyMp
        The ground state upon which the excitation was based
    amplitude : AmplitudeVector
        The amplitude vector
    intermediates : adcc.Intermediates
        Intermediates from the ADC calculation to reuse
    """
    if not isinstance(method, IsrMethod):
        method = IsrMethod(method)
    if not isinstance(ground_state, LazyMp):
        raise TypeError("ground_state should be a LazyMp object.")
    if not isinstance(amplitude, AmplitudeVector):
        raise TypeError("amplitude should be an AmplitudeVector object.")
    if intermediates is None:
        intermediates = Intermediates(ground_state)

    if method.name not in DISPATCH:
        raise NotImplementedError("state_diffdm is not implemented "
                                  f"for {method.name}.")
    else:
        ret = DISPATCH[method.name](ground_state, amplitude, intermediates)
        return ret.evaluate()



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2020 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import numpy as np
from scipy import constants
import warnings

from .ElectronicStates import ElectronicStates, _timer_name
from .Excitation import Excitation
from .misc import cached_member_function
from .OneParticleOperator import OneParticleOperator
from .NParticleOperator import product_trace


class ElectronicTransition(ElectronicStates):
    # The child classes S2S and ExcitedStates currently
    # both share the Excitation class for the state view.
    # This might change in the future once e.g. ExcitedStates
    # specific methods are implemented.
    _state_view_cls = Excitation

    @property
    def excitations(self) -> list[Excitation]:
        """
        Provides a list of Excitations, i.e., a view to all individual
        excitations and their properties.
        """
        return [self._state_view(i) for i in range(self.size)]

    @property
    def transition_dm(self) -> list[OneParticleOperator]:
        """List of transition density matrices of all computed states"""
        return [self._transition_dm(i) for i in range(self.size)]

    @cached_member_function(timer=_timer_name, separate_timings_by_args=False)
    def _transition_dm(self, state_n: int) -> OneParticleOperator:
        """Computes the tansition density matrix for a single state"""
        evec = self.excitation_vector[state_n]
        return self._module.transition_dm(
            self.property_method, self.ground_state, evec, self.matrix.intermediates
        )

    @property
    def transition_dipole_moment(self) -> np.ndarray:
        """Array of transition dipole moments of all computed states"""
        return np.array([
            self._transition_dipole_moment(i) for i in range(self.size)
        ])

    @cached_member_function(timer=_timer_name, separate_timings_by_args=False)
    def _transition_dipole_moment(self, state_n: int) -> np.ndarray:
        """Computes the transition dipole moment for a single state"""
        if self.property_method.level == 0:
            warnings.warn("ADC(0) transition dipole moments are known to be "
                          "faulty in some cases.")
        dipole_integrals = self.operators.electric_dipole
        tdm = self._transition_dm(state_n)
        return np.array(
            [product_trace(comp, tdm) for comp in dipole_integrals]
        )

    @property
    def transition_dipole_moment_velocity(self) -> np.ndarray:
        """
        Array of transition dipole moments in the velocity gauge of all
        computed states
        """
        return np.array([
            self._transition_dipole_moment_velocity(i) for i in range(self.size)
        ])

    @cached_member_function(timer=_timer_name, separate_timings_by_args=False)
    def _transition_dipole_moment_velocity(self, state_n: int) -> np.ndarray:
        """
        Computes the transition dipole moments in the velocity gauge for a
        single state
        """
        if self.property_method.level == 0:
            warnings.warn("ADC(0) transition velocity dipole moments "
                          "are known to be faulty in some cases.")
        dipole_integrals = self.operators.electric_dipole_velocity
        tdm = self._transition_dm(state_n)
        return np.array(
            [product_trace(comp, tdm) for comp in dipole_integrals]
        )

    def transition_magnetic_dipole_moment(self,
                                          gauge_origin="origin") -> np.ndarray:
        """Array of transition magnetic dipole moments of all computed states"""
        return np.array([
            self._transition_magnetic_dipole_moment(state_n=i,
                                                    gauge_origin=gauge_origin)
            for i in range(self.size)
        ])

    @cached_member_function(timer=_timer_name, separate_timings_by_args=False)
    def _transition_magnetic_dipole_moment(self, state_n: int,
                                           gauge_origin="origin") -> np.ndarray:
        """
        Computes the transition magnetic dipole moments for a single state
        """
        if self.property_method.level == 0:
            warnings.warn("ADC(0) transition magnetic dipole moments "
                          "are known to be faulty in some cases.")
        mag_dipole_integrals = self.operators.magnetic_dipole(gauge_origin)
        tdm = self._transition_dm(state_n)
        return np.array([
            product_trace(comp, tdm) for comp in mag_dipole_integrals
        ])

    def transition_quadrupole_moment(self, gauge_origin="origin") -> np.ndarray:
        """Array of transition quadrupole moments of all computed states"""
        return np.array([
            self._transition_quadrupole_moment(state_n=i, gauge_origin=gauge_origin)
            for i in range(self.size)
        ])

    @cached_member_function(timer=_timer_name, separate_timings_by_args=False)
    def _transition_quadrupole_moment(self, state_n: int, gauge_origin="origin"):
        """Computes the transition quadrupole moments for a single state"""
        if self.property_method.level == 0:
            warnings.warn("ADC(0) transition quadrupole moments are known to be "
                          "faulty in some cases.")
        quadrupole_integrals = self.operators.electric_quadrupole(gauge_origin)
        tdm = self._transition_dm(state_n)
        return np.array([
            [product_trace(q, tdm) for q in quad] for quad in quadrupole_integrals
        ])

    def transition_quadrupole_moment_velocity(self,
                                              gauge_origin="origin") -> np.ndarray:
        """Array of transition quadrupole moments of all computed states"""
        return np.array([
            self._transition_quadrupole_moment_velocity(state_n=i,
                                                        gauge_origin=gauge_origin)
            for i in range(self.size)
        ])

    @cached_member_function(timer=_timer_name, separate_timings_by_args=False)
    def _transition_quadrupole_moment_velocity(self, state_n: int,
                                               gauge_origin="origin") -> np.ndarray:
        """Compute the transition quadrupole moments for a single state"""
        if self.property_method.level == 0:
            warnings.warn("ADC(0) transition velocity quadrupole moments are known "
                          "to be faulty in some cases.")
        quadrupole_integrals = (
            self.operators.electric_quadrupole_velocity(gauge_origin)
        )
        tdm = self._transition_dm(state_n)
        return np.array([
            [product_trace(q, tdm) for q in quad] for quad in quadrupole_integrals
        ])

    @property
    def oscillator_strength(self) -> np.ndarray:
        """Array of oscillator strengths of all computed states"""
        return np.array([self._oscillator_strength(i) for i in range(self.size)])

    @cached_member_function(timer=_timer_name, separate_timings_by_args=False)
    def _oscillator_strength(self, state_n: int) -> np.float64:
        """Computes the oscillator strengths for a single state"""
        tdm = self._transition_dipole_moment(state_n)
        ev = self.excitation_energy[state_n]
        return 2. / 3. * np.linalg.norm(tdm)**2 * np.abs(ev)

    @property
    def oscillator_strength_velocity(self) -> np.ndarray:
        """Array of oscillator strengths in velocity gauge of all computed states"""
        return np.array([
            self._oscillator_strength_velocity(i) for i in range(self.size)
        ])

    @cached_member_function(timer=_timer_name, separate_timings_by_args=False)
    def _oscillator_strength_velocity(self, state_n: int) -> np.float64:
        """Computes the oscillator strength in velocity gauge for a single state"""
        tdm = self._transition_dipole_moment_velocity(state_n)
        ev = self.excitation_energy[state_n]
        return 2. / 3. * np.linalg.norm(tdm)**2 / np.abs(ev)

    @property
    def rotatory_strength(self) -> np.ndarray:
        """
        Array of rotatory strengths (in velocity gauge) of all computed states.
        This property is gauge-origin invariant, thus, it is not possible to
        select a gauge origin.
        """
        return np.array([
            self._rotatory_strength(i) for i in range(self.size)
        ])

    @cached_member_function(timer=_timer_name, separate_timings_by_args=False)
    def _rotatory_strength(self, state_n: int) -> np.float64:
        """
        Computes the rotatory strength (in velocity gauge) for a single state.
        This property is gauge-origin invariant, thus, it is not possible to
        select a gauge origin.
        """
        tdm = self._transition_dipole_moment_velocity(state_n)
        magmom = self._transition_magnetic_dipole_moment(
            state_n=state_n, gauge_origin="origin"
        )
        ee = self.excitation_energy[state_n]
        return np.dot(tdm, magmom) / ee

    def rotatory_strength_length(self, gauge_origin="origin") -> np.ndarray:
        """Array of rotatory strengths in length gauge of all computed states"""
        return np.array([
            self._rotatory_strength_length(state_n=i, gauge_origin=gauge_origin)
            for i in range(self.size)
        ])

    @cached_member_function(timer=_timer_name, separate_timings_by_args=False)
    def _rotatory_strength_length(self, state_n: int,
                                  gauge_origin="origin") -> np.float64:
        """Computes the rotatory strength in length gauge for a single state"""
        tdm = self._transition_dipole_moment(state_n)
        magmom = self._transition_magnetic_dipole_moment(state_n=state_n,
                                                         gauge_origin=gauge_origin)
        return -1.0 * np.dot(tdm, magmom)

    @property
    def cross_section(self) -> np.ndarray:
        """Array of one-photon absorption cross sections of all computed states"""
        return np.array([self._cross_section(i) for i in range(self.size)])

    def _cross_section(self, state_n: int) -> np.float64:
        """Computes the one-photon absorption cross sections for a single state"""
        # TODO Source?
        fine_structure = constants.fine_structure
        fine_structure_au = 1 / fine_structure
        prefac = 2.0 * np.pi ** 2 / fine_structure_au
        return prefac * self._oscillator_strength(state_n)

    def plot_spectrum(self, broadening="lorentzian", xaxis="eV",
                      yaxis="cross_section", width=0.01,
                      width_unit: str = "au", **kwargs):
        """
        One-shot plotting function for the spectrum generated by all states
        known to this class.

        Makes use of the :class:`adcc.visualisation.Spectrum` class
        in order to generate and format the spectrum to be plotted, using
        many sensible defaults.

        Parameters
        ----------
        broadening : str or None or callable, optional
            The broadening type to used for the computed excitations.
            A value of None disables broadening any other value is passed
            straight to
            :func:`adcc.visualisation.Spectrum.broaden_lines`.
        xaxis : str
            Energy unit to be used on the x-Axis. Options:
            ["eV", "au", "nm", "cm-1"]
        yaxis : str
            Quantity to plot on the y-Axis. Options are "cross_section",
            "osc_strength", "dipole" (plots norm of transition dipole),
            "rotational_strength" (ECD spectrum with rotational strength)
        width : float, optional
            Gaussian broadening standard deviation or Lorentzian broadening
            gamma parameter. The value should be given in atomic units
            and will be converted to the unit of the energy axis.
        width_unit: str, optional
            The unit the width is given in. All xaxis options except "nm" are
            possible.
        """
        if yaxis in ["osc", "osc_strength", "oscillator_strength", "f"]:
            yvalues = self.oscillator_strength
            ylabel = "Oscillator strengths (au)"
        elif yaxis in ["dipole", "dipole_norm", "μ"]:
            yvalues = np.linalg.norm(self.transition_dipole_moment, axis=1)
            ylabel = "Modulus of transition dipole (au)"
        elif yaxis in ["cross_section", "σ"]:
            yvalues = self.cross_section
            ylabel = "Cross section (au)"
        elif yaxis in ["rot", "rotational_strength", "rotatory_strength"]:
            yvalues = self.rotatory_strength
            ylabel = "Rotatory strength (au)"
        else:
            raise ValueError(f"Unknown yaxis specifier: {yaxis}")

        return self._plot_spectrum(
            yvalues=yvalues, ylabel=ylabel, xaxis=xaxis, broadening=broadening,
            width=width, width_unit=width_unit, **kwargs
        )



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2018 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import numpy as np

from .misc import cached_property, cached_member_function
from .Tensor import Tensor
from .timings import Timer, timed_member_call
from .OneParticleOperator import OneParticleOperator
from .OneParticleDensity import OneParticleDensity
from .NParticleOperator import OperatorSymmetry, NParticleOperator
from .TwoParticleOperator import TwoParticleOperator
from .functions import einsum
from .MoSpaces import split_spaces

import libadcc


def transform_operator_ao2mo(tensor_bb: Tensor, tensor_ff: NParticleOperator,
                             coefficients,
                             conv_tol: float = 1e-14):
    """Take a block-diagonal tensor in the atomic orbital basis
    and transform it into the molecular orbital basis in the
    convention used by adcc.

    Parameters
    ----------
    tensor_bb : Tensor
        Block-diagonal tensor in the atomic orbital basis
    tensor_ff : Tensor
        Output tensor with the symmetry set-up to contain
        the operator in the molecular orbital representation
    coefficients : callable
        Function providing coefficient blocks
    conv_tol : float, optional
        SCF convergence tolerance, by default 1e-14
    """
    for blk in tensor_ff.canonical_blocks:
        if len(blk) == 4:
            cleft = coefficients(blk[:2] + "b")
            cright = coefficients(blk[2:] + "b")
            temp = cleft @ tensor_bb @ cright.transpose()

            # TODO: once the permutational symmetry is correct:
            # tensor_ff.set_block(blk, tensor_ff)
            tensor_ff[blk].set_from_ndarray(temp.to_ndarray(), conv_tol)

        elif len(blk) == 8:
            cleft_1 = coefficients(blk[:2] + "b")
            cleft_2 = coefficients(blk[2:4] + "b")
            cright_1 = coefficients(blk[4:6] + "b")
            cright_2 = coefficients(blk[6:] + "b")
            temp = einsum("ia,jb,abcd,kc,ld->ijkl",
                          cleft_1, cleft_2, tensor_bb, cright_1, cright_2)

            # TODO: once the permutational symmetry is correct:
            # tensor_ff.set_block(blk, tensor_ff)
            tensor_ff[blk].set_from_ndarray(temp.to_ndarray(), conv_tol)
        else:
            raise NotImplementedError(
                "Only one- and two-particle operators are implemented."
            )


def transform_operator_ao2mo_spin_projected(tensor_bb: Tensor,
                                            tensor_ff: NParticleOperator,
                                            coeff_map: dict[str, Tensor],
                                            spin_map: str = "aa",
                                            conv_tol: float = 1e-14):
    """Take a tensor in the atomic orbital basis
    and transform it into the molecular orbital basis in the
    convention used by adcc.

    The transformation is performed block-wise using the provided
    molecular orbital coefficient matrices for the selected
    spin components.

    Parameters
    ----------
    tensor_bb : Tensor
        Tensor in the atomic orbital basis
    tensor_ff : Tensor
        Output tensor with the symmetry set-up to contain
        the operator in the molecular orbital representation
    coeff_map : dict
        Dictionary containing molecular orbital coefficient matrices,
        keyed by orbital space and spin label (e.g. "<space>_a", "<space>_b").
    spin_map : str, optional
        Two-character string specifying which spin components are projected
        for the left and right indices (e.g. "aa", "ab"). Default is "aa".
    conv_tol : float, optional
        SCF convergence tolerance, by default 1e-14
    """
    assert len(spin_map) == 2
    spin1, spin2 = list(spin_map)

    for blk in tensor_ff.canonical_blocks:
        if len(blk) == 4:
            s1, s2 = split_spaces(blk)
            cleft = coeff_map[f"{s1}_{spin1}"]
            cright = coeff_map[f"{s2}_{spin2}"]
            temp = cleft @ tensor_bb @ cright.transpose()

            # TODO: once the permutational symmetry is correct:
            # tensor_ff.set_block(blk, tensor_ff)
            tensor_ff[blk].set_from_ndarray(temp.to_ndarray(), conv_tol)
        else:
            raise NotImplementedError


def replicate_ao_block(mospaces, tensor,
                       symmetry: OperatorSymmetry = OperatorSymmetry.HERMITIAN,
                       block: str = "ab"):
    """
    transform_operator_ao2mo requires the operator in the AO basis to be
    replicated in a block-diagonal fashion (e.g. for a OneParticleOperator:
    [A  0
     0  A]).
    This is achieved using this function.

    The `block` argument controls which blocks are constructed:
    - block="ab": replicate the operator for both alpha and beta spaces,
      resulting in a full block-diagonal structure.
    - block="a": construct only the corresponding single block.
    """
    assert block in ["ab", "a"]
    zerobk = np.zeros_like(tensor)
    if len(tensor.shape) == 2:
        sym = libadcc.make_symmetry_operator_basis(
            mospaces, tensor.shape[0], symmetry.to_str(), 1, block
        )
        result = Tensor(sym)

        if block == "ab":
            result.set_from_ndarray(np.block([
                [tensor, zerobk],
                [zerobk, tensor],
            ]), 1e-14)
        else:
            result.set_from_ndarray(np.block([
                tensor
            ]), 1e-14)
    elif len(tensor.shape) == 4:
        sym = libadcc.make_symmetry_operator_basis(
            mospaces, tensor.shape[0], symmetry.to_str(), 2, block
        )
        result = Tensor(sym)
        if block == "ab":
            tensor_ex = - tensor.transpose((0, 1, 3, 2))
            tensor_as = tensor + tensor_ex
            full_tensor = np.block([
                [
                    [  # [aaaa, aaab], [aaba aabb]
                        [tensor_as, zerobk],
                        [zerobk, zerobk],
                    ],
                    [  # [abaa, abab], [abba, abbb]
                        [zerobk, tensor],
                        [tensor_ex, zerobk],
                    ],
                ],
                [
                    [  # [baaa, baab], [baba, babb]
                        [zerobk, tensor_ex],
                        [tensor, zerobk],
                    ],
                    [  # [bbaa, bbab], [bbba, bbbb]
                        [zerobk, zerobk],
                        [zerobk, tensor_as],
                    ],
                ],
            ])
            result.set_from_ndarray(full_tensor, 1e-14)
        else:
            raise NotImplementedError(
                f"Invalid block {block}. Only 'ab' supported for 2-particle "
                "operators."
            )
    else:
        raise NotImplementedError("Only 2 and 4 dimensional tensors supported.")
    return result


class OperatorIntegrals:
    def __init__(self, provider, mospaces, coefficients, coefficients_alpha,
                 coefficients_beta, conv_tol):
        self._provider_ao = provider
        self.mospaces = mospaces
        self._coefficients = coefficients
        self._coefficients_alpha = coefficients_alpha
        self._coefficients_beta = coefficients_beta
        self._conv_tol = conv_tol
        self._import_timer = Timer()

    @property
    def provider_ao(self):
        """
        The data structure which provides the integral data in the
        atomic orbital basis from the backend.
        """
        return self._provider_ao

    @property
    def available(self) -> tuple[str, ...]:
        """Which integrals are available in the underlying backend"""
        return self.provider_ao.available

    @cached_property
    @timed_member_call("_import_timer")
    def overlap_ao(self) -> Tensor:
        """Return the overlap in the atomic orbital basis."""
        if "overlap" not in self.available:
            raise NotImplementedError(f"overlap operator not implemented "
                                      f"in {self.provider_ao.backend} backend.")

        ao_operator = getattr(self.provider_ao, "overlap")
        ovlp_bb = replicate_ao_block(self.mospaces, ao_operator,
                                     symmetry=OperatorSymmetry.HERMITIAN,
                                     block="a")
        return ovlp_bb

    @cached_property
    @timed_member_call("_import_timer")
    def ssq_1p(self) -> OneParticleOperator:
        """Returns the one-particle part of the S^2 operator"""
        op = OneParticleOperator(self.mospaces, symmetry=OperatorSymmetry.HERMITIAN)
        # d_ij = 3/4 \delta_ij
        for ss in self.mospaces.subspaces_occupied:
            op[ss + ss].set_mask("ii", 0.75)
        # d_ab = 3/4 \delta_ij
        for ss in self.mospaces.subspaces_virtual:
            op[ss + ss].set_mask("ii", 0.75)
        return op

    @cached_property
    @timed_member_call("_import_timer")
    def ssq_2p(self) -> TwoParticleOperator:
        """Returns the two-particle part of the S^2 operator"""
        # currently not implemented for CVS
        if any(sp not in ["o1", "v1"] for sp in self.mospaces.subspaces):
            raise NotImplementedError("The 2-particle part of the SSq operator is "
                                      "only implemented for the occupied and "
                                      "virtual spaces, i.e., "
                                      "CVS is not supported yet.")
        # Intermediates
        # S^aa, S^ab and S^bb (spin projected overlap matrices)
        ovlp_bb: Tensor = self.overlap_ao
        coeff_map = {}
        for sp in self.mospaces.subspaces:
            coeff_map[sp + "_a"] = self._coefficients_alpha(sp + "b")
            coeff_map[sp + "_b"] = self._coefficients_beta(sp + "b")

        S_aa = OneParticleOperator(self.mospaces,
                                   symmetry=OperatorSymmetry.NOSYMMETRY)
        transform_operator_ao2mo_spin_projected(ovlp_bb, S_aa, coeff_map, "aa",
                                                self._conv_tol)

        S_ab = OneParticleOperator(self.mospaces,
                                   symmetry=OperatorSymmetry.NOSYMMETRY)
        transform_operator_ao2mo_spin_projected(ovlp_bb, S_ab, coeff_map, "ab",
                                                self._conv_tol)

        S_bb = OneParticleOperator(self.mospaces,
                                   symmetry=OperatorSymmetry.NOSYMMETRY)
        transform_operator_ao2mo_spin_projected(ovlp_bb, S_bb, coeff_map, "bb",
                                                self._conv_tol)

        # additional intermediate
        S_aa_minus_bb = S_aa - S_bb

        op = TwoParticleOperator(self.mospaces, symmetry=OperatorSymmetry.HERMITIAN)
        op.oooo = (
            + 1.0 * einsum("ik,jl->ijkl", S_aa_minus_bb.oo,  S_aa_minus_bb.oo)
            # exploit symmetry S_ab.oo = S_ba.oo.T
            + 4.0 * einsum("ik,jl->ijkl", S_ab.oo.T, S_ab.oo)
        ).antisymmetrise(2, 3)
        op.ooov = (
            + 2.0 * einsum("ik,ja->ijka", S_ab.oo.T, S_ab.ov)
            + 2.0 * einsum("ik,ja->ijka", S_ab.oo, S_ab.vo.T)
        ).antisymmetrise(0, 1)
        op.oovv = (
            + 2.0 * einsum("ia,jb->ijab", S_ab.vo.T, S_ab.ov)
            + 2.0 * einsum("ia,jb->ijab", S_ab.ov, S_ab.vo.T)
        ).antisymmetrise(2, 3)
        op.ovov = (
            + 0.5 * einsum("ij,ab->iajb", S_aa_minus_bb.oo, S_aa_minus_bb.vv)
            + 1.0 * einsum("ij,ab->iajb", S_ab.oo.T, S_ab.vv)
            + 1.0 * einsum("ij,ab->iajb", S_ab.oo, S_ab.vv.T)
            - 1.0 * einsum("ib,aj->iajb", S_ab.vo.T, S_ab.vo)
            - 1.0 * einsum("ib,aj->iajb", S_ab.ov, S_ab.ov.T)
        )
        op.ovvv = (
            + 2.0 * einsum("ib,ac->iabc", S_ab.vo.T, S_ab.vv)
            + 2.0 * einsum("ib,ac->iabc", S_ab.ov, S_ab.vv.T)
        ).antisymmetrise(2, 3)
        op.vvvv = (
            + 1.0 * einsum("ac,bd->abcd", S_aa_minus_bb.vv, S_aa_minus_bb.vv)
            # exploit symmetry S_ab.vv = S_ba.vv.T
            + 4.0 * einsum("ac,bd->abcd", S_ab.vv, S_ab.vv.T)
        ).antisymmetrise(2, 3)
        return op

    def _import_dipole_like_operator(
        self, integral: str,
        symmetry: OperatorSymmetry = OperatorSymmetry.HERMITIAN
    ) -> tuple[OneParticleOperator, ...]:
        if integral not in self.available:
            raise NotImplementedError(f"{integral.replace('_', ' ')} operator "
                                      "not implemented "
                                      f"in {self.provider_ao.backend} backend.")

        ao_operator = getattr(self.provider_ao, integral)
        assert len(ao_operator) == 3  # has to have a x, y and z component

        dipoles = []
        for comp in range(3):  # [x, y, z]
            dip_bb = replicate_ao_block(self.mospaces, ao_operator[comp],
                                        symmetry=symmetry)
            dip_ff = OneParticleOperator(self.mospaces, symmetry=symmetry)
            transform_operator_ao2mo(dip_bb, dip_ff, self._coefficients,
                                     self._conv_tol)
            dipoles.append(dip_ff)
        return tuple(dipoles)

    @cached_property
    @timed_member_call("_import_timer")
    def electric_dipole(self) -> tuple[OneParticleOperator, ...]:
        """Return the electric dipole integrals in the molecular orbital basis."""
        return self._import_dipole_like_operator(
            "electric_dipole", symmetry=OperatorSymmetry.HERMITIAN)

    @cached_property
    @timed_member_call("_import_timer")
    def electric_dipole_velocity(self) -> tuple[OneParticleOperator, ...]:
        """
        Return the electric dipole integrals (in the velocity gauge)
        in the molecular orbital basis.
        """
        return self._import_dipole_like_operator(
            "electric_dipole_velocity", symmetry=OperatorSymmetry.ANTIHERMITIAN)

    def _import_g_origin_dep_dip_like_operator(
            self, integral: str, gauge_origin="origin",
            symmetry: OperatorSymmetry = OperatorSymmetry.HERMITIAN
    ) -> tuple[OneParticleOperator, ...]:
        """
        Imports the operator and transforms it to the molecular orbital basis.

        Parameters
        ----------
        integral : str
            The dipole like gauge dependent integral to import: an integral
            that consists of 3 components (x, y, z) and whose AO import function
            takes the gauge origin as argument.
        gauge_origin: str or tuple[str]
            The gauge origin used for the generation of the AO integrals.
        is_symmetric : bool, optional
            If the imported operator is symmetric, by default True
        """
        if integral not in self.available:
            raise NotImplementedError(f"{integral} operator is not implemented "
                                      f"in {self.provider_ao.backend} backend.")

        ao_operator = getattr(self.provider_ao, integral)(gauge_origin)
        assert len(ao_operator) == 3  # has to have a x, y and z component

        dipoles = []
        for comp in range(3):  # [x, y, z]
            dip_bb = replicate_ao_block(self.mospaces, ao_operator[comp],
                                        symmetry=symmetry)
            dip_ff = OneParticleOperator(self.mospaces, symmetry=symmetry)
            transform_operator_ao2mo(dip_bb, dip_ff, self._coefficients,
                                     self._conv_tol)
            dipoles.append(dip_ff)
        return tuple(dipoles)

    # separate the timings, so one can easily see in the timings how many different
    # gauge_origins were used throughout the calculation
    @cached_member_function(timer="_import_timer", separate_timings_by_args=True)
    def magnetic_dipole(self, gauge_origin="origin"
                        ) -> tuple[OneParticleOperator, ...]:
        """
        Returns the magnetic dipole intergrals
        in the molecular orbital basis dependent on the selected gauge origin.
        The default gauge origin is set to (0.0, 0.0, 0.0) (= 'origin').
        """
        return self._import_g_origin_dep_dip_like_operator(
            integral="magnetic_dipole", gauge_origin=gauge_origin,
            symmetry=OperatorSymmetry.ANTIHERMITIAN
        )

    def _import_g_origin_dep_quad_like_operator(
        self, integral: str, gauge_origin="origin",
        symmetry: OperatorSymmetry = OperatorSymmetry.HERMITIAN
    ) -> tuple[tuple[OneParticleOperator, ...], ...]:  # noqa E501
        """
        Imports the operator and transforms it to the molecular orbital basis.

        Parameters
        ----------
        integral : str
            The quadrupole like gauge dependent integral to import: an integral
            that consists of 9 components (xx, xy, xz, ... zz)
            and whose AO import function takes the gauge origin as single argument.
        gauge_origin: str or tuple[str]
            The gauge origin used for the generation of the AO integrals.
        is_symmetric : bool, optional
            if the imported operator is symmetric, by default True
        """
        if integral not in self.available:
            raise NotImplementedError(f"{integral} operator is not implemented "
                                      f"in {self.provider_ao.backend} backend.")

        ao_operator = getattr(self.provider_ao, integral)(gauge_origin)
        assert len(ao_operator) == 9

        flattened = []
        for comp in range(9):  # [xx, xy, xz, yx, yy, yz, zx, zy, zz]
            quad_bb = replicate_ao_block(self.mospaces, ao_operator[comp],
                                         symmetry=symmetry)
            quad_ff = OneParticleOperator(
                self.mospaces, symmetry=symmetry
            )
            transform_operator_ao2mo(quad_bb, quad_ff, self._coefficients,
                                     self._conv_tol)
            flattened.append(quad_ff)
        return (tuple(flattened[:3]), tuple(flattened[3:6]), tuple(flattened[6:]))

    @cached_member_function(timer="_import_timer", separate_timings_by_args=True)
    def electric_quadrupole(self, gauge_origin="origin"
                            ) -> tuple[tuple[OneParticleOperator, ...], ...]:
        """
        Returns the electric quadrupole integrals
        in the molecular orbital basis dependent on the selected gauge origin.
        The default gauge origin is set to (0.0, 0.0, 0.0) (= 'origin').
        """
        return self._import_g_origin_dep_quad_like_operator(
            integral="electric_quadrupole", gauge_origin=gauge_origin,
            symmetry=OperatorSymmetry.HERMITIAN
        )

    @cached_member_function(timer="_import_timer", separate_timings_by_args=True)
    def electric_quadrupole_traceless(self, gauge_origin="origin"
                                      ) -> tuple[tuple[OneParticleOperator, ...], ...]:  # noqa E501
        """
        Returns the traceless electric quadrupole integrals
        in the molecular orbital basis dependent on the selected gauge origin.
        The default gauge origin is set to (0.0, 0.0, 0.0) (= 'origin').
        """
        return self._import_g_origin_dep_quad_like_operator(
            integral="electric_quadrupole_traceless", gauge_origin=gauge_origin,
            symmetry=OperatorSymmetry.HERMITIAN
        )

    @cached_member_function(timer="_import_timer", separate_timings_by_args=True)
    def electric_quadrupole_velocity(self, gauge_origin="origin"
                                     ) -> tuple[tuple[OneParticleOperator, ...], ...]:  # noqa E501
        """
        Returns the electric quadrupole integrals in velocity gauge
        in the molecular orbital basis dependent on the selected gauge origin.
        The default gauge origin is set to (0.0, 0.0, 0.0) (= 'origin').
        """
        return self._import_g_origin_dep_quad_like_operator(
            integral="electric_quadrupole_velocity", gauge_origin=gauge_origin,
            symmetry=OperatorSymmetry.ANTIHERMITIAN
        )

    @cached_member_function(timer="_import_timer", separate_timings_by_args=True)
    def diamagnetic_magnetizability(self, gauge_origin="origin"
                                    ) -> tuple[tuple[OneParticleOperator, ...], ...]:  # noqa E501
        """
        Returns the diamagnetic magnetizability integrals
        in the molecular orbital basis dependent on the selected gauge origin.
        The default gauge origin is set to (0.0, 0.0, 0.0) (= 'origin').
        """
        return self._import_g_origin_dep_quad_like_operator(
            integral="diamagnetic_magnetizability", gauge_origin=gauge_origin,
            symmetry=OperatorSymmetry.HERMITIAN
        )

    def _import_density_dependent_operator(
        self, operator: str, density_mo: OneParticleDensity,
        symmetry: OperatorSymmetry = OperatorSymmetry.HERMITIAN
    ) -> OneParticleDensity:
        """
        Import the density-dependent operator and transform it to the
        molecular orbital basis.

        Parameters
        ----------
        integral : str
            The density-dependent operator to import: an operator
            whose AO import function takes a density matrix as single argument.
        density_mo: OneParticleDensity
            The density in the MO basis for which to compute the operator.
        is_symmetric : bool, optional
            if the imported operator is symmetric, by default True
        """
        dm_ao = sum(density_mo.to_ao_basis())
        v_ao = getattr(self.provider_ao, operator)(dm_ao)
        v_bb = replicate_ao_block(
            self.mospaces, v_ao, symmetry=symmetry
        )
        v_ff = OneParticleDensity(self.mospaces, symmetry=symmetry)
        transform_operator_ao2mo(
            v_bb, v_ff, self._coefficients, self._conv_tol
        )
        return v_ff

    def pe_induction_elec(self,
                          density_mo: OneParticleDensity) -> OneParticleDensity:
        """
        Returns the (density-dependent) PE electronic induction operator in the
        molecular orbital basis.
        """
        if "pe_induction_elec" not in self.available:
            raise NotImplementedError("PE electronic induction operator "
                                      "not implemented "
                                      f"in {self.provider_ao.backend} backend.")
        return self._import_density_dependent_operator(
            operator="pe_induction_elec", density_mo=density_mo,
            symmetry=OperatorSymmetry.HERMITIAN
        )

    def pcm_potential_elec(self,
                           density_mo: OneParticleDensity) -> OneParticleDensity:
        """
        Returns the (density-dependent) electronic PCM potential operator in the
        molecular orbital basis
        """
        if "pcm_potential_elec" not in self.available:
            raise NotImplementedError("Electronic PCM potential operator "
                                      "not implemented "
                                      f"in {self.provider_ao.backend} backend.")
        return self._import_density_dependent_operator(
            operator="pcm_potential_elec", density_mo=density_mo,
            symmetry=OperatorSymmetry.HERMITIAN
        )

    def _import_2p_like_operator(
        self, integral: str,
        symmetry: OperatorSymmetry = OperatorSymmetry.HERMITIAN
    ) -> TwoParticleOperator:
        if integral not in self.available:
            raise NotImplementedError(f"{integral.replace('_', ' ')} operator "
                                      "not implemented "
                                      f"in {self.provider_ao.backend} backend.")

        ao_operator = getattr(self.provider_ao, integral)

        op_bbbb = replicate_ao_block(self.mospaces, ao_operator,
                                     symmetry=symmetry)

        op_ffff = TwoParticleOperator(self.mospaces, symmetry=symmetry)
        transform_operator_ao2mo(op_bbbb, op_ffff, self._coefficients,
                                 self._conv_tol)
        return op_ffff

    @property
    def timer(self):
        ret = Timer()
        ret.attach(self._import_timer, subtree="import")
        return ret



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2020 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import numpy as np

from . import adc_pp
from .AdcMethod import AdcType
from .ElectronicStates import _timer_name
from .ElectronicTransition import ElectronicTransition
from .misc import cached_member_function
from .OneParticleOperator import OneParticleOperator


class State2States(ElectronicTransition):
    def __init__(self, data, method=None, property_method=None, initial=0):
        """Construct a State2States class from some data obtained
            from an interative solver or an :class:`ExcitedStates` object.

            The class provides access to ADC transition properties between
            excited states, i.e., from the `initial` state to all higher-lying
            excited states obtained from an ADC calculation.

            By default the ADC method is extracted from the data object
            and the property method in property_method is set equal to
            this method, except ADC(3) where property_method=="adc2".
            This can be overwritten using the parameters.

            Parameters
            ----------
            data
                Any kind of iterative solver state. Typically derived off
                a :class:`solver.EigenSolverStateBase`. Can also be an
                :class:`ExcitedStates` object.
            method : str, optional
                Provide an explicit method parameter if data contains none.
            property_method : str, optional
                Provide an explicit method for property calculations to
                override the automatic selection.
            initial : int, optional
                Provide the index of the excited state from which transitions
                to all other higher-lying states are to be computed.
        """
        super().__init__(data, method, property_method)
        self.initial = initial

        # Since this class should be used for all adc_types we have to determine
        # the module according to the method.
        if self.method.adc_type is AdcType.PP:
            self._module = adc_pp
        else:
            raise ValueError(f"Unknown adc_type {self.method.adc_type}.")

    @property
    def size(self) -> int:
        # the number of states "above" the initial state to correctly index
        # into the transition_dm array
        return super().size - self.initial - 1

    @property
    def excitation_energy(self) -> np.ndarray:
        """
        Excitation energies from the inital state to energetically higher lying
        states in atomic units
        """
        return np.array([
            self._excitation_energy[final]
            - self._excitation_energy[self.initial]
            for final in range(self.initial + 1, super().size)
        ])

    @property
    def excitation_energy_uncorrected(self) -> np.ndarray:
        """
        Excitation energies without any corrections from the inital state to
        energetically higher lying states in atomic units
        """
        return np.array([
            self._excitation_energy_uncorrected[final]
            - self._excitation_energy_uncorrected[self.initial]
            for final in range(self.initial + 1, super().size)
        ])

    @property
    def excitation_vector(self):
        """List of excitation vectors"""
        return self._excitation_vector[self.initial + 1:]

    @property
    def transition_dm(self) -> list[OneParticleOperator]:
        """
        List of transition density matrices from initial state to final state/s
        """
        return [self._transition_dm(final) for final in range(self.size)]

    @cached_member_function(timer=_timer_name, separate_timings_by_args=False)
    def _transition_dm(self, state_n: int) -> OneParticleOperator:
        """
        Computes the transition density matrices from initial state to a single
        final state
        """
        # NOTE: state_n is relative to the initial state, i.e.,
        # 0 refers to initial_state + 1.
        # This is necessary to enable the use of the implementations on
        # the parent class.
        state_n = self.initial + state_n + 1
        return self._module.state2state_transition_dm(
            self.property_method, self.ground_state,
            super().excitation_vector[self.initial],
            super().excitation_vector[state_n],
            self.matrix.intermediates
        )

    def _state_diffdm(self, state_n: int) -> OneParticleOperator:
        """
        The difference density matrix are not available through the
        :class:`State2States`
        class. Please use e.g. :class:`ExcitedStates` (for PP-ADC) instead
        """
        # alternatively one could create an independent StateProperties class
        # and additionally inherit from that class on ExcitedStates and the
        # future IP/EA class.
        raise NotImplementedError("Difference density matrices and excited state "
                                  "properties are not available through "
                                  f"'{self.__class__.__name__}'. Please use e.g. "
                                  "'ExcitedStates' for PP-ADC.")



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
import adcc

from pyscf import gto, scf
from matplotlib import pyplot as plt

# Run SCF in pyscf
mol = gto.M(
    atom='O 0 0 0;'
         'H 0 0 1.795239827225189;'
         'H 1.693194615993441 0 -0.599043184453037',
    basis='cc-pvdz',
    unit="Bohr"
)
scfres = scf.RHF(mol)
scfres.conv_tol = 1e-13
scfres.kernel()

# Run an adc3 calculation:
singlets     = adcc.adc3(scfres,                   n_singlets=3)  # noqa: E221
singlets_fc  = adcc.adc3(scfres, frozen_core=1,    n_singlets=3)  # noqa: E221
singlets_fv2 = adcc.adc3(scfres, frozen_virtual=2, n_singlets=3)  # noqa: E221
singlets_fv4 = adcc.adc3(scfres, frozen_virtual=4, n_singlets=3)  # noqa: E221
singlets_fv6 = adcc.adc3(scfres, frozen_virtual=6, n_singlets=3)  # noqa: E221
singlets_fv8 = adcc.adc3(scfres, frozen_virtual=8, n_singlets=3)  # noqa: E221
singlets.plot_spectrum(label="ADC(3)")
singlets_fc.plot_spectrum(label="FC-ADC(3)")
singlets_fv2.plot_spectrum(label="FV-ADC(3) 2")
singlets_fv4.plot_spectrum(label="FV-ADC(3) 4")
singlets_fv6.plot_spectrum(label="FV-ADC(3) 6")
singlets_fv8.plot_spectrum(label="FV-ADC(3) 8")

plt.legend()
plt.show()



import numpy as np
from scipy import constants
import warnings

from .AdcMatrix import AdcMatrix
from .AdcMethod import AdcMethod, IsrMethod, MethodLevel
from .AmplitudeVector import AmplitudeVector
from .FormatDominantElements import FormatDominantElements
from .FormatIndex import (
    FormatIndexBase, FormatIndexAdcc, FormatIndexHfProvider, FormatIndexHomoLumo
)
from .LazyMp import LazyMp
from .OneParticleDensity import OneParticleDensity
from .TwoParticleDensity import TwoParticleDensity
from .NParticleOperator import product_trace
from .OperatorIntegrals import OperatorIntegrals
from .misc import cached_member_function, requires_module
from .ReferenceState import ReferenceState
from .solver.SolverStateBase import EigenSolverStateBase
from .StateView import StateView
from .timings import Timer
from .visualisation.Spectrum import Spectrum


_timer_name = "_property_timer"


class ElectronicStates:
    # The module where the equations for the ADC scheme are implemented, e.g.,
    # adc_pp for PP-ADC. Overwrite the class variable on child classes to
    # forward the call to the corresponding module. This assumes that the
    # structure within each module is consistent for all adc_types, which should
    # be fine I think.
    _module = None
    # The class used to obtain a view on a single state, e.g., Excitation for the
    # ExcitedStates class. Has to be defined on the corresponding child class.
    _state_view_cls: StateView = None

    def __init__(self, data, method: str = None,
                 property_method: str = None) -> None:
        """
        Construct an ElectronicStates class from some data obtained from an
        iterative solver or another :class:`ElectronicStates` instance.

        Parameters
        ----------
        data
            Any kind of iterative solver state. Typically derived off
            a :class:`solver.EigenSolverStateBase`.
        method : str, optional
            Provide an explicit method parameter if data contains none.
        property_method : str, optional
            Provide an explicit method for property calculations to
            override the automatic selection.
    """
        self.matrix: AdcMatrix = data.matrix
        self.ground_state: LazyMp = self.matrix.ground_state
        self.reference_state: ReferenceState = (
            self.matrix.ground_state.reference_state
        )
        self.operators: OperatorIntegrals = self.reference_state.operators

        # List of all the objects which have timers (do not yet collect
        # timers, since new times might be added implicitly at a later point)
        self._property_timer: Timer = Timer()
        self._timed_objects = [("", self.reference_state),
                               ("adcmatrix", self.matrix),
                               ("mp", self.ground_state),
                               ("intermediates", self.matrix.intermediates)]
        if hasattr(data, "timer"):
            datakey = getattr(data, "algorithm", data.__class__.__name__)
            self._timed_objects.append((datakey, data))

        # Copy some optional attributes
        for optattr in ["converged", "spin_change", "kind", "n_iter"]:
            if hasattr(data, optattr):
                setattr(self, optattr, getattr(data, optattr))

        self.method = getattr(data, "method", method)
        if self.method is None:
            self.method = self.matrix.method
        if not isinstance(self.method, AdcMethod):
            self.method: AdcMethod = AdcMethod(self.method)

        if property_method is None:
            if self.method.level in [MethodLevel.TWO_X, MethodLevel.THREE]:
                # Auto-select ISR(2) properties for ADC(2)-x and ADC(3) calc
                warnings.warn(f"ISR({self.method.level.to_str()}) not implemented."
                              f" Property method is selected as ISR(2).")
                property_method = self.method.at_level(2).as_method(IsrMethod)
            else:
                property_method = self.method.as_method(IsrMethod)
        elif not isinstance(property_method, IsrMethod):
            property_method = IsrMethod(property_method)
        self._property_method: IsrMethod = property_method

        # Special stuff for special solvers
        if isinstance(data, EigenSolverStateBase):
            self._excitation_vector = data.eigenvectors
            self._excitation_energy_uncorrected = data.eigenvalues
            self.residual_norm = data.residual_norms
        else:  # other Solver or e.g. from another ExcitedStates instance
            if hasattr(data, "eigenvalues"):
                self._excitation_energy_uncorrected = data.eigenvalues
            if hasattr(data, "eigenvectors"):
                self._excitation_vector = data.eigenvectors
            # if both excitation_energy and excitation_energy_uncorrected
            # are present, the latter one has priority: otherwise corrections
            # would be added again below!
            if hasattr(data, "excitation_energy"):
                self._excitation_energy_uncorrected = \
                    data.excitation_energy.copy()
            if hasattr(data, "excitation_energy_uncorrected"):
                self._excitation_energy_uncorrected =\
                    data.excitation_energy_uncorrected.copy()
            if hasattr(data, "excitation_vector"):
                self._excitation_vector = data.excitation_vector

        self._excitation_energy = self._excitation_energy_uncorrected.copy()

        # Collect all excitation energy corrections
        self._excitation_energy_corrections: list[EnergyCorrection] = []
        # copy energy corrections if possible
        # and avoids re-computation of the corrections
        if hasattr(data, "_excitation_energy_corrections"):
            for eec in data._excitation_energy_corrections:
                correction_energy = getattr(data, eec.name)
                if hasattr(self, eec.name):
                    raise ValueError(f"{self.__name__} already has an attribute "
                                     f"with the name '{eec.name}'")
                setattr(self, eec.name, correction_energy)
                self._excitation_energy += correction_energy
                self._excitation_energy_corrections.append(eec)

    def __len__(self) -> int:
        return self.size

    @property
    def size(self) -> int:
        return self._excitation_energy.size

    @property
    def property_method(self) -> IsrMethod:
        """The method used to evaluate ISR properties"""
        return self._property_method

    @property
    def timer(self) -> Timer:
        """Return a cumulative timer collecting timings from the calculation"""
        ret = Timer()
        for key, obj in self._timed_objects:
            ret.attach(obj.timer, subtree=key)
        ret.attach(self._property_timer, subtree="properties")
        ret.time_construction = self.reference_state.timer.time_construction
        return ret

    @property
    def excitation_energy(self):
        """Excitation energies including all corrections in atomic units"""
        return self._excitation_energy

    @property
    def excitation_energy_uncorrected(self):
        """Excitation energies without any corrections in atomic units"""
        return self._excitation_energy_uncorrected

    @property
    def excitation_vector(self):
        """List of excitation vectors"""
        return self._excitation_vector

    @property
    def state_diffdm(self) -> list[OneParticleDensity]:
        """List of difference density matrices of all computed states"""
        return [self._state_diffdm(state_n=i) for i in range(self.size)]

    @cached_member_function(timer=_timer_name, separate_timings_by_args=False)
    def _state_diffdm(self, state_n: int) -> OneParticleDensity:
        """Computes the difference density matrix for a single state"""
        evec = self.excitation_vector[state_n]
        return self._module.state_diffdm(
            self.property_method, self.ground_state, evec, self.matrix.intermediates
        )

    @property
    def state_dm(self) -> list[OneParticleDensity]:
        """List of state density matrices of all computed states"""
        return [self._state_dm(i) for i in range(self.size)]

    def _state_dm(self, state_n: int) -> OneParticleDensity:
        """List of state density matrices of all computed states"""
        mp_density = self.ground_state.density(self.property_method.level.to_int())
        diffdm = self._state_diffdm(state_n)
        return mp_density + diffdm

    @property
    def state_diffdm_2p(self) -> list[TwoParticleDensity]:
        """
        List of two-particle difference density matrices of all computed states
        """
        return [self._state_diffdm_2p(state_n=i) for i in range(self.size)]

    @cached_member_function(timer=_timer_name, separate_timings_by_args=False)
    def _state_diffdm_2p(self, state_n: int) -> TwoParticleDensity:
        """Computes the two-particle difference density matrix for a single state"""
        evec = self.excitation_vector[state_n]
        return self._module.state_diffdm_2p(
            self.property_method, self.ground_state, evec, self.matrix.intermediates
        )

    @property
    def state_dm_2p(self) -> list[TwoParticleDensity]:
        """List of two-particle state density matrices of all computed states"""
        return [self._state_dm_2p(i) for i in range(self.size)]

    def _state_dm_2p(self, state_n: int) -> TwoParticleDensity:
        """List of two particle state density matrices of all computed states"""
        mp_density = self.ground_state.density_2p(
            self.property_method.level.to_int()
        )
        diffdm = self._state_diffdm_2p(state_n)
        return mp_density + diffdm

    @property
    def state_ssq(self) -> np.ndarray:
        """Array of state <S^2>"""
        return np.array([self._state_ssq(i) for i in range(self.size)])

    @cached_member_function(timer=_timer_name, separate_timings_by_args=False)
    def _state_ssq(self, state_n: int) -> float:
        """Computes the <S^2> of a single state."""
        pmethod = self.property_method
        if pmethod.level.to_int() == 0:
            gs_ssq = self.reference_state.ssq
        else:
            gs_ssq = self.ground_state.ssq(pmethod.level.to_int())

        ssq_1p_op = self.operators.ssq_1p
        ssq_2p_op = self.operators.ssq_2p
        ddm_1p = self._state_diffdm(state_n)
        ddm_2p = self._state_diffdm_2p(state_n)
        ssq_1p = product_trace(ssq_1p_op, ddm_1p)
        ssq_2p = product_trace(ssq_2p_op, ddm_2p)
        return (ssq_1p + ssq_2p + gs_ssq)

    @property
    def state_dipole_moment(self) -> np.ndarray:
        """Array of state dipole moments"""
        return np.array([self._state_dipole_moment(i) for i in range(self.size)])

    @cached_member_function(timer=_timer_name, separate_timings_by_args=False)
    def _state_dipole_moment(self, state_n: int) -> np.ndarray:
        """Computes the state dipole moment for a single state"""
        pmethod = self.property_method
        if pmethod.level.to_int() == 0:
            gs_dip_moment = self.reference_state.dipole_moment
        else:
            gs_dip_moment = self.ground_state.dipole_moment(pmethod.level.to_int())

        dipole_integrals = self.operators.electric_dipole
        ddm = self._state_diffdm(state_n)
        return (gs_dip_moment
                + np.array([product_trace(comp, ddm) for comp in dipole_integrals]))

    def _state_view(self, state_n: int) -> StateView:
        """
        Provides a view onto a single state and his properties.
        """
        return self._state_view_cls(self, state_n)

    def _add_energy_correction(self, correction: "EnergyCorrection") -> None:
        assert isinstance(correction, EnergyCorrection)
        if hasattr(self, correction.name):
            raise ValueError(f"{self.__class__.__name__} already has an attribute "
                             f" with the name '{correction.name}'")
        correction_energy = np.array([
            correction(self._state_view(i)) for i in range(self.size)
        ])
        setattr(self, correction.name, correction_energy)
        self._excitation_energy += correction_energy
        self._excitation_energy_corrections.append(correction)

    def __iadd__(self, other):
        if isinstance(other, EnergyCorrection):
            self._add_energy_correction(other)
        elif isinstance(other, list):
            for k in other:
                self += k
        else:
            return NotImplemented
        return self

    def __add__(self, other):
        if not isinstance(other, (EnergyCorrection, list)):
            return NotImplemented
        ret = self.__class__(self, self.method, self.property_method)
        ret += other
        return ret

    @property
    def excitation_property_keys(self) -> list[str]:
        """
        Extracts the names of available excited state and transition properties.
        """
        # NOTE: this currently assumes that all available properties are available
        # on the corresponding state_view class
        assert self._state_view_cls is not None
        blacklist = ("parent_state")
        ret = []
        for key in dir(self._state_view_cls):
            # private fields or ao transformed densities or other fields
            if key.startswith("_") or key.endswith("_ao") or key in blacklist:
                continue
            ret.append(key)
        return ret

    @requires_module("pandas")
    def to_dataframe(self, gauge_origin="origin"):
        """
        Exports the class object as :class:`pandas.DataFrame`.
        Atomic units are used for all values.
        """
        import pandas as pd
        propkeys = self.excitation_property_keys
        propkeys.extend([k.name for k in self._excitation_energy_corrections])
        data = {
            "excitation": np.arange(0, self.size, dtype=int),
            "kind": np.tile(self.kind, self.size)
        }
        for key in propkeys:
            try:
                d = getattr(self, key)
            except NotImplementedError:
                # some properties are not available for every backend
                continue
            # This assumes that all callables take a single argument:
            # the gauge origin
            if callable(d):
                try:
                    d = d(gauge_origin)
                except NotImplementedError:
                    # some properties are not available for every backend
                    continue

            if not isinstance(d, np.ndarray):
                continue
            if not np.issubdtype(d.dtype, np.number):
                continue

            if d.ndim == 1:
                data[key] = d
            elif d.ndim == 2 and d.shape[1] == 3:
                for i, p in enumerate(["x", "y", "z"]):
                    data[f"{key}_{p}"] = d[:, i]
            elif d.ndim == 3 and d.shape[1:] == (3, 3):
                for i, p in enumerate(["x", "y", "z"]):
                    for j, q in enumerate(["x", "y", "z"]):
                        data[f"{key}_{p}{q}"] = d[:, i, j]
            elif d.ndim > 3:
                warnings.warn(f"Exporting NumPy array for property {key}"
                              f" with shape {d.shape} not supported.")
                continue
        df = pd.DataFrame(data=data)
        df.set_index("excitation")
        return df

    @requires_module("matplotlib")
    def _plot_spectrum(self, ylabel: str, yvalues: np.ndarray, xaxis: str = "eV",
                       broadening="lorentzian", width: float = 0.01,
                       width_unit: str = "au", **kwargs):
        """
        One-shot plotting function for the spectrum generated by all states
        known to this class.

        Makes use of the :class:`adcc.visualisation.Spectrum` class in order to
        generate and format the spectrum to be plotted,
        using many sensible defaults.

        Parameters
        ----------
        ylabel: str
            The label for the y-Axis.
        yvalues: np.ndarray
            The y-values to plot.
        xaxis : str
            Energy unit to be used on the x-Axis. Options:
            ["eV", "au", "nm", "cm-1"].
        broadening : str or None or callable, optional
            The broadening type to used for the computed excitations.
            A value of None disables broadening any other value is passed
            straight to
            :func:`adcc.visualisation.Spectrum.broaden_lines`.
        width : float, optional
            Gaussian broadening standard deviation or Lorentzian broadening
            gamma parameter. The value should be given in atomic units
            and will be converted to the unit of the energy axis.
        width_unit: str, optional
            The unit the width is given in. All xaxis options except "nm" are
            possible.
        """

        def convert_x_units(spectrum: Spectrum):
            # modifies the spectrum in place by converting the x values to
            # the desired unit
            if xaxis in ["au", "hartree", "a.u."]:
                spectrum.xlabel = "Energy (au)"
            elif xaxis == "ev":
                eV = constants.value("Hartree energy in eV")
                spectrum.x = spectrum.x * eV
                spectrum.xlabel = "Energy (eV)"
            elif xaxis == "nm":
                hc = constants.h * constants.c
                Eh = constants.value("Hartree energy")
                spectrum.x = hc / (spectrum.x * Eh) * 1e9
                spectrum.xlabel = "Wavelength (nm)"
            elif xaxis in ["cm-1", "cm^-1", "cm^{-1}"]:
                towvn = constants.value("hartree-inverse meter relationship") / 100
                spectrum.x = spectrum.x * towvn
                spectrum.xlabel = "Wavenumbers (cm^{-1})"
            else:
                raise ValueError(f"Unknown xaxis specifier: {xaxis}")

        # convert the width to au
        width_unit = width_unit.lower()
        if width_unit in ["au", "hartree", "a.u."]:
            pass
        elif width_unit == "ev":
            width = width / constants.value("Hartree energy in eV")
        elif width_unit in ["cm-1", "cm^-1", "cm^{-1}"]:
            towvn = constants.value("hartree-inverse meter relationship") / 100
            width = width / towvn
        else:
            raise ValueError(f"Invalid width_unit '{width_unit}'")

        # first build the spectra in atomic units on the x-axis
        spectrum = Spectrum(x=self.excitation_energy, y=yvalues, ylabel=ylabel)
        if broadening:  # generate a broadened spectrum
            broadened_spectrum: Spectrum = spectrum.broaden_lines(
                width=width, shape=broadening
            )
        xaxis = xaxis.lower()
        # convert the x-unit of the dicrete spectrum
        convert_x_units(spectrum)
        if broadening:
            # convert the x-unit of the broadened spectrum
            convert_x_units(broadened_spectrum)
            # plot both spectra:
            # - avoid labeling the discrete spectrum
            kwdisc = {k: v for k, v in kwargs.items() if k != "label"}
            plots = spectrum.plot(style="discrete", **kwdisc)
            # - use the same color as for the discrete spectrum
            kwbroad = {k: v for k, v in kwargs.items() if k != "color"}
            disc_color = plots[0].get_color()
            plots.extend(broadened_spectrum.plot(
                color=disc_color, style="continuous", **kwbroad
            ))
        else:
            # plot the discrete spectrum
            plots = spectrum.plot(style="discrete", **kwargs)
        return plots

    def _describe(self, columns: list["TableColumn"]):
        """
        Return a string providing a human-readable description of the class

        Parameters
        ----------
        columns: list[TableColumn]
            The columns of the table that is generated by the method.
        """
        lines = []
        # - Determine the overall width of the table
        table_width = sum(c.width for c in columns)
        # the width of the header will be taken account below
        # but we also need to consider the excitation energy corrections
        # that are printed below after
        if self._excitation_energy_corrections:
            correction_head = "Excitation energy includes these corrections:"
            corr_width = max(len(correction_head),
                             *(len(c.name) for c in
                               self._excitation_energy_corrections))
            if corr_width > table_width:
                # add padding to the last column
                new_width = columns[-1].width + corr_width - table_width
                columns[-1] = columns[-1].with_width(new_width)
                table_width = corr_width

        # - Format the header
        # Format the method
        method = self.method.name
        if self.property_method != self.method:
            method += f" ({self.property_method.name})"
        # Format the state information: kind, spin_change and convergence
        state_info = []
        if hasattr(self, "kind") and self.kind:
            state_info.append(self.kind)
        if hasattr(self, "spin_change") and self.spin_change is not None and \
                self.spin_change != 0:
            state_info.append(f"(ΔMS={self.spin_change:+2d})")
        if hasattr(self, "converged"):
            conv = "converged" if self.converged else "NOT CONVERGED"
            if state_info:  # add separator to previous entry
                state_info[-1] += ","
            state_info.append(conv)
        state_info = " ".join(state_info)
        # actually format the header
        if table_width > len(method) + 4:
            header = (  # -4 for the spaces
                f"{method:s}    {state_info:>{str(table_width - len(method) - 4)}s}"
            )
        else:  # only few small columns -> header will determine the table width
            header = f"{method:s}    {state_info:s}"

        # if the header is wider than the current table width:
        # add padding to the last column
        if len(header) > table_width:
            new_width = columns[-1].width + len(header) - table_width
            columns[-1] = columns[-1].with_width(new_width)
            table_width = len(header)
        assert len(header) == table_width

        # The hline separator for the table:
        # +2: add a space to the left and the right
        table_hline = "+" + "-" * (table_width + 2) + "+"

        # Actually format the table
        lines.append(table_hline)
        lines.append(f"| {header} |")
        lines.append(table_hline)
        lines.append(f"| {''.join(c.header for c in columns)} |")
        lines.append(f"| {''.join(c.unit for c in columns)} |")
        assert all(len(c.values) == self.size for c in columns)
        for state_values in zip(*[c.values for c in columns]):
            lines.append(f"| {''.join(state_values)} |")
        lines.append(table_hline)
        # Potentially format energy corrections
        if self._excitation_energy_corrections:
            assert len(correction_head) <= table_width
            lines.append(f"| {correction_head:<{str(table_width)}s} |")
            maxlen = table_width - 5  # -5 for the spaces and - below
            for eec in self._excitation_energy_corrections:
                lines.append(f"|    - {eec.name:<{maxlen}s} |")
        return "\n".join(lines)

    def describe_amplitudes(self, tolerance: float = 0.01,
                            index_format=None) -> str:
        """
        Return a string describing the dominant amplitudes of each
        excitation vector in human-readable form. The ``kwargs``
        are for :py:class:`FormatExcitationVector`.

        Parameters
        ----------
        tolerance : float, optional
            Minimal absolute value of the excitation amplitudes considered.

        index_format : NoneType or str or FormatIndexBase, optional
            Formatter to use for displaying tensor indices.
            Valid are ``"adcc"`` to keep the adcc-internal indexing,
            ``"hf"`` to select the HFProvider indexing, ``"homolumo"``
            to index relative on the HOMO / LUMO / HOCO orbitals.
            If ``None`` an automatic selection will be made.
        """
        eV = constants.value("Hartree energy in eV")
        formatter = FormatAmplitudeVector(
            matrix=self.matrix, tolerance=tolerance, index_format=index_format
        )
        # optimise the formatting by pre-inspecting the amplitude coefficients
        # to determine the (str) length of indices that will be printed.
        formatter.optimise_formatting(tuple(self.excitation_vector))
        # first format all the headers to determine their length
        headers = []
        for i, energy in enumerate(self.excitation_energy):
            headers.append(
                f"State {i:3d}, {energy:13.7g} au, {energy * eV:13.7g} eV"
            )
        # determine the required linewidth
        width = max(formatter.linewidth, *(len(h) for h in headers))
        hline = "+" + (width + 2) * "-" + "+"  # +2: space to the left and right

        lines = []
        for header, amplitude in zip(headers, self.excitation_vector):
            lines.append(hline)
            lines.append(f"| {header:<{width}s} |")
            lines.append(hline)
            # format the vector and add the table borders + spaces
            for coefficient in formatter.format(amplitude):
                lines.append(
                    f"| {coefficient:^{width}s} |"
                )
            lines.append(hline)
        return "\n".join(lines)


class EnergyCorrection:
    def __init__(self, name: str, function: callable) -> None:
        """A helper class to represent excitation energy
        corrections.

        Parameters
        ----------
        name : str
            descriptive name of the energy correction
        function : callable
            function that takes a :py:class:`Excitation`
            as single argument and returns the energy
            correction as a float
        """
        if not isinstance(name, str):
            raise TypeError("name needs to be a string.")
        if not callable(function):
            raise TypeError("function needs to be callable.")
        self.name: str = name
        self.function: callable = function

    def __call__(self, excitation: StateView):
        assert isinstance(excitation, StateView)
        return self.function(excitation)


class TableColumn:
    # represents a column in the table generated by the describe method
    def __init__(self, header: str, values: tuple[str, ...], unit: str = "",
                 width: int = None, align: str = "^"):
        # the minimum required width of the column
        min_width = max(len(header), len(unit), *(len(v) for v in values))
        if width is None or width < min_width:
            width = min_width + 1

        self.width = width
        self.header = f"{header:{align}{str(width)}s}"
        self.values = tuple(f"{v:{align}{str(width)}s}" for v in values)
        self.unit = f"{unit:{align}{str(width)}s}"
        self.align = align

    def with_width(self, width: int) -> "TableColumn":
        # remove whitespaces in case we want to shrink the column!
        values = tuple(v.strip() for v in self.values)
        return TableColumn(
            header=self.header.strip(), values=values, unit=self.unit.strip(),
            width=width, align=self.align
        )

    def with_align(self, align: str) -> "TableColumn":
        # remove whitespaces so we can change the alignment
        values = tuple(v.strip() for v in self.values)
        return TableColumn(
            header=self.header.strip(), values=values, unit=self.unit.strip(),
            width=self.width, align=align
        )


class FormatAmplitudeVector:
    def __init__(self, matrix: AdcMatrix, tolerance: float = 0.01,
                 index_format: str = None):
        """
        Set up a formatter class for formatting excitation vectors.

        Parameters
        ----------
        tolerance : float, optional
            Minimal absolute value of the excitation amplitudes considered

        index_format : NoneType or str or FormatIndexBase, optional
            Formatter to use for displaying tensor indices.
            Valid are ``"adcc"`` to keep the adcc-internal indexing,
            ``"hf"`` to select the HFProvider indexing, ``"homolumo"``
            to index relative on the HOMO / LUMO / HOCO orbitals.
            If ``None`` an automatic selection will be made.
        """
        self.matrix = matrix
        refstate = matrix.reference_state
        # default for index formatting
        if index_format is None:
            closed_shell = refstate.n_alpha == refstate.n_beta
            if closed_shell and refstate.is_aufbau_occupation:
                index_format = "homolumo"
            else:
                index_format = "hf"

        if index_format == "adcc":
            index_format = FormatIndexAdcc(refstate)
        elif index_format == "hf":
            index_format = FormatIndexHfProvider(refstate)
        elif index_format == "homolumo":
            index_format = FormatIndexHomoLumo(refstate)
        elif not isinstance(index_format, FormatIndexBase):
            raise ValueError("Unsupported value for index_format: "
                             + str(index_format))

        self.tensor_format = FormatDominantElements(
            matrix.mospaces, tolerance, index_format
        )
        self.index_format = self.tensor_format.index_format
        # Formatting used for the values (coefficients)
        self.value_format = "{:+8.3g}"

    def optimise_formatting(self, vectors: list[AmplitudeVector]) -> None:
        """
        Optimise the formatting of the amplitude vectors by inspecting the entries
        of the given amplitude vectors in advance.
        This enables to nicely format indices by e.g. determining the maximum
        length of indices that will be printed.
        """
        if not isinstance(vectors, (list, tuple)):
            return self.optimise_formatting([vectors])

        for vector in vectors:
            for block, spaces in self.matrix.axis_spaces.items():
                self.tensor_format.optimise_formatting((spaces, vector[block]))

    @property
    def linewidth(self) -> int:
        """
        The width of a line if an amplitude is formatted with this class
        """
        # This assumes a formatting: idx1 idx2 -> idx3 idx4   aa->aa   number
        nblk = len(self.matrix.axis_blocks[-1])  # axis blocks is sorted by len
        # n_idx * (len(index) + space) + 2 for the arrow
        width_indices = nblk * (self.index_format.max_n_characters + 1) + 2
        width_spins = nblk + 2  # n_idx * 1 (for a/b) + 2 for the arrow
        # the width of the numerical value of the coefficient
        width_value = len(self.value_format.format(0))
        # +6: 3 spaces between indices and spin + 3 spaces between spins and value
        return width_indices + width_spins + width_value + 6

    def format(self, amplitude: AmplitudeVector) -> list[str]:
        """
        Formats the given amplitude vector.
        """
        empty_idx = self.index_format.max_n_characters * " "
        idx_spin_gap = " " * 3
        spin_coeff_gap = " " * 3

        if self.matrix.axis_blocks == ["ph"]:
            formats = {"ov": (
                "{} -> {}" + idx_spin_gap + "{}->{}"
                + spin_coeff_gap + self.value_format
            )}
        elif self.matrix.axis_blocks == ["ph", "pphh"]:
            formats = {
                "ov": (
                    "{} " + empty_idx + " -> {} " + empty_idx + idx_spin_gap
                    + "{} ->{} " + spin_coeff_gap + self.value_format
                ),
                "oovv": (
                    "{} {} -> {} {}" + idx_spin_gap + "{}{}->{}{}"
                    + spin_coeff_gap + self.value_format
                )
            }
        else:
            raise NotImplementedError("Unknown ADC matrix structure")

        lines = []
        for block, spaces in self.matrix.axis_spaces.items():
            # remove the numbers from the spaces: o1 -> o
            space_str = "".join(c for c in "".join(spaces) if c.isalpha())

            assert block in amplitude
            formatted = self.tensor_format.format_as_list(spaces, amplitude[block])
            for indices, spins, coeff in formatted:
                lines.append(
                    formats[space_str].format(*indices, *spins, coeff)
                )
        return lines



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2020 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
from .state_diffdm import state_diffdm
from .state_diffdm_2p import state_diffdm_2p
from .transition_dm import transition_dm
from .state2state_transition_dm import state2state_transition_dm
from .modified_transition_moments import modified_transition_moments

"""
Submodule, which contains rather lengthy low-level kernels
(e.g. matrix-vector products or working equations), which are called
from the high-level objects in the adcc main module.
"""

__all__ = ["state_diffdm", "state_diffdm_2p", "state2state_transition_dm",
           "transition_dm", "modified_transition_moments"]



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
import adcc

from import_data import import_data

# Gather precomputed data
data = import_data()

# Make it unrestricted
data["restricted"] = False

# Run an unrestricted adc2 calculation:
states = adcc.adc2(data, n_states=10, conv_tol=1e-8)
print(states.describe())



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2019 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import libadcc
import numpy as np

from .functions import direct_sum, evaluate, einsum, zeros_like
from .misc import cached_property, cached_member_function
from .ReferenceState import ReferenceState
from .OneParticleDensity import OneParticleDensity
from .TwoParticleDensity import TwoParticleDensity
from .NParticleOperator import product_trace, OperatorSymmetry
from .Intermediates import register_as_intermediate
from .timings import Timer, timed_member_call
from .MoSpaces import MoSpaces, split_spaces
from . import block as b


class LazyMp:
    def __init__(self, hf):
        """
        Initialise the class dealing with the Møller-Plesset ground state.
        """
        if isinstance(hf, libadcc.HartreeFockSolution_i):
            hf = ReferenceState(hf)
        if not isinstance(hf, ReferenceState):
            raise TypeError("hf needs to be a ReferenceState "
                            "or a HartreeFockSolution_i")
        self.reference_state: ReferenceState = hf
        self.mospaces: MoSpaces = hf.mospaces
        self.timer: Timer = Timer()
        self.has_core_occupied_space: bool = hf.has_core_occupied_space

    def __getattr__(self, attr):
        # Shortcut some quantities, which are needed most often
        if attr.startswith("t2") and len(attr) == 4:  # t2oo, t2oc, t2cc
            xxvv = b.__getattr__(attr[2:4] + "vv")
            return self.t2(xxvv)
        else:
            raise AttributeError

    @cached_member_function()
    def df(self, space):
        """Delta Fock matrix"""
        hf = self.reference_state
        s1, s2 = split_spaces(space)
        fC = hf.fock(s1 + s1).diagonal()
        fv = hf.fock(s2 + s2).diagonal()
        return direct_sum("-i+a->ia", fC, fv)

    @cached_member_function()
    def t2(self, space):
        """T2 amplitudes"""
        hf = self.reference_state
        sp = split_spaces(space)
        assert all(s == b.v for s in sp[2:])
        eia = self.df(sp[0] + b.v)
        ejb = self.df(sp[1] + b.v)
        return (
            hf.eri(space) / direct_sum("ia+jb->ijab", eia, ejb).symmetrise((2, 3))
        )

    @cached_member_function()
    def td2(self, space):
        """Return the T^D_2 term"""
        if space != b.oovv:
            raise NotImplementedError("T^D_2 term not implemented "
                                      f"for space {space}.")
        t2erit = self.t2eri(b.oovv, b.ov).transpose((1, 0, 2, 3))
        denom = direct_sum(
            'ia,jb->ijab', self.df(b.ov), self.df(b.ov)
        ).symmetrise(0, 1)
        return (
            + 4.0 * t2erit.antisymmetrise(2, 3).antisymmetrise(0, 1)
            - 0.5 * self.t2eri(b.oovv, b.vv)
            - 0.5 * self.t2eri(b.oovv, b.oo)
        ) / denom

    @cached_member_function()
    def tt2(self, space: str):
        """
        Return the second order MP triples amplitudes for the given space
        (e.g. o1o1o1v1v1v1).
        """
        if space != b.ooovvv:
            raise NotImplementedError("Second order MP triples amplitudes not "
                                      f"implemented for space {space}.")
        hf = self.reference_state
        df = self.df(b.ov)
        t2_1 = self.t2(b.oovv)
        # denom = a + b + c - i - j - k   //   df = i - a
        denom = - 1 * direct_sum(
            "ia,jkbc->ijkabc", df, direct_sum("jb,kc->jkbc", df, df)
        ).symmetrise(0, 1, 2).symmetrise(3, 4, 5)
        # prefactor of 9, because we have 9 terms each in the expression, while the
        # antisymmetrisation generates 36 terms and introduces a prefactor of 1/36.
        # Each of the 9 terms is therefore generated 4 times.
        # The scaling in the comments is given as: [comp_scaling] / [mem_scaling]
        numerator = (
            # N^7: O^3V^4 / N^6: O^3V^3
            + 9 * einsum('idab,jkcd->ijkabc', hf.ovvv, t2_1)
            # N^7: O^4V^3 / N^6: O^3V^3
            + 9 * einsum('ijla,klbc->ijkabc', hf.ooov, t2_1)
        ).antisymmetrise(0, 1, 2).antisymmetrise(3, 4, 5)
        return numerator / denom

    @cached_member_function()
    def t2eri(self, space, contraction):
        """
        Return the T2 tensor with ERI tensor contraction intermediates.
        These are called pi1 to pi7 in libadc.
        """
        hf = self.reference_state
        key = space + contraction
        expressions = {
            # space + contraction
            b.ooov + b.vv: ('ijbc,kabc->ijka', b.ovvv),
            b.ooov + b.ov: ('ilab,lkjb->ijka', b.ooov),
            b.oovv + b.oo: ('klab,ijkl->ijab', b.oooo),
            b.oovv + b.ov: ('jkac,kbic->ijab', b.ovov),
            b.oovv + b.vv: ('ijcd,abcd->ijab', b.vvvv),
            b.ovvv + b.oo: ('jkbc,jkia->iabc', b.ooov),
            b.ovvv + b.ov: ('ijbd,jcad->iabc', b.ovvv),
        }
        if key not in expressions:
            raise NotImplementedError("t2eri intermediate not implemented "
                                      f"for space '{space}' and contraction "
                                      f"'{contraction}'.")
        contraction_str, eri_block = expressions[key]
        return einsum(contraction_str, self.t2oo, hf.eri(eri_block))

    @cached_property
    @timed_member_call(timer="timer")
    def mp2_diffdm(self):
        """
        Return the MP2 difference density in the MO basis.
        """
        hf = self.reference_state
        ret = OneParticleDensity(self.mospaces,
                                 symmetry=OperatorSymmetry.HERMITIAN)
        # NOTE: the following 3 blocks are equivalent to the cvs_p0 intermediates
        # defined at the end of this file
        ret.oo = -0.5 * einsum("ikab,jkab->ij", self.t2oo, self.t2oo)
        ret.ov = -0.5 * (
            + einsum("ijbc,jabc->ia", self.t2oo, hf.ovvv)
            + einsum("jkib,jkab->ia", hf.ooov, self.t2oo)
        ) / self.df(b.ov)
        ret.vv = 0.5 * einsum("ijac,ijbc->ab", self.t2oo, self.t2oo)

        if self.has_core_occupied_space:
            # additional terms to "revert" CVS for ground state density
            ret.oo += -0.5 * einsum("iLab,jLab->ij", self.t2oc, self.t2oc)
            ret.ov += -0.5 * (
                + einsum("jMib,jMab->ia", hf.ocov, self.t2oc)
                + einsum("iLbc,Labc->ia", self.t2oc, hf.cvvv)
                + einsum("kLib,kLab->ia", hf.ocov, self.t2oc)
                + einsum("iMLb,LMab->ia", hf.occv, self.t2cc)
                - einsum("iLMb,LMab->ia", hf.occv, self.t2cc)
            ) / self.df(b.ov)
            ret.vv += (
                + 0.5 * einsum("IJac,IJbc->ab", self.t2cc, self.t2cc)
                + 1.0 * einsum("kJac,kJbc->ab", self.t2oc, self.t2oc)
            )
            # compute extra CVS blocks
            ret.cc = -0.5 * (
                + einsum("kIab,kJab->IJ", self.t2oc, self.t2oc)
                + einsum('LIab,LJab->IJ', self.t2cc, self.t2cc)
            )
            ret.oc = -0.5 * (
                + einsum("kIab,kjab->jI", self.t2oc, self.t2oo)
                + einsum("ILab,jLab->jI", self.t2cc, self.t2oc)
            )
            ret.cv = -0.5 * (
                - einsum("jIbc,jabc->Ia", self.t2oc, hf.ovvv)
                + einsum("jkIb,jkab->Ia", hf.oocv, self.t2oo)
                + einsum("jMIb,jMab->Ia", hf.occv, self.t2oc)
                + einsum("ILbc,Labc->Ia", self.t2cc, hf.cvvv)
                + einsum("kLIb,kLab->Ia", hf.occv, self.t2oc)
                + einsum("LMIb,LMab->Ia", hf.cccv, self.t2cc)
            ) / self.df(b.cv)
        ret.reference_state = self.reference_state
        return evaluate(ret)

    def density(self, level=2):
        """
        Return the MP density in the MO basis with all corrections
        up to the specified order of perturbation theory
        """
        if level in [0, 1]:
            return self.reference_state.density
        elif level == 2:
            return self.reference_state.density + self.mp2_diffdm
        else:
            raise NotImplementedError("Only densities for level 0, 1 and 2"
                                      " are implemented.")

    @cached_property
    @timed_member_call(timer="timer")
    def mp1_dm_correction_2p(self) -> TwoParticleDensity:
        """
        Return the two-particle MP1 difference density correction in the MO basis.
        """
        ret = TwoParticleDensity(self.mospaces,
                                 symmetry=OperatorSymmetry.HERMITIAN)
        ret.oovv = -1.0 * self.t2oo
        return ret.evaluate()

    @cached_property
    @timed_member_call(timer="timer")
    def mp2_dm_correction_2p(self) -> TwoParticleDensity:
        """
        Return the two-particle MP2 difference density correction in the MO basis.
        """
        hf: ReferenceState = self.reference_state
        ret = TwoParticleDensity(self.mospaces,
                                 symmetry=OperatorSymmetry.HERMITIAN)
        p0: OneParticleDensity = self.mp2_diffdm

        # constuct Kronecker Delta
        d_oo = zeros_like(hf.foo)
        d_oo.set_mask("ii", 1)

        ret.oooo = (
            + 4.0 * einsum("ik,jl->ijkl", p0.oo, d_oo)
            .antisymmetrise(0, 1).antisymmetrise(2, 3)
            + 0.5 * einsum("ijab,klab->ijkl", self.t2oo, self.t2oo)
        )
        ret.ooov = (
            + 2.0 * einsum("ja,ik->ijka", p0.ov, d_oo).antisymmetrise(0, 1)
        )
        ret.oovv = (
            - 1.0 * self.td2(b.oovv)
        )
        ret.ovov = (
            + 1.0 * einsum("ab,ij->iajb", p0.vv, d_oo)
            - 1.0 * einsum("jkac,ikbc->iajb", self.t2oo, self.t2oo)
        )
        ret.vvvv = (
            + 0.5 * einsum("ijab,ijcd->abcd", self.t2oo, self.t2oo)
        )
        return evaluate(ret)

    def diffdm_2p(self, level=2) -> TwoParticleDensity:
        """
        Return the two-particle MP difference density in the MO basis with all
        corrections up to the specified order of perturbation theory.
        """
        if level == 1:
            return self.mp1_dm_correction_2p
        elif level == 2:
            return (self.mp1_dm_correction_2p
                    + self.mp2_dm_correction_2p)
        else:
            raise NotImplementedError("Only first and second-order two-particle "
                                      "density corrections are implemented.")

    def density_2p(self, level=2) -> TwoParticleDensity:
        """
        Return the two-particle MP density in the MO basis with all corrections
        up to the specified order of perturbation theory.
        """
        if level == 0:
            return self.reference_state.density_2p
        diffdm = self.diffdm_2p(level)
        return self.reference_state.density_2p + diffdm

    def dipole_moment(self, level=2):
        """
        Return the MP dipole moment at the specified level of
        perturbation theory.
        """
        if level in [0, 1]:
            return self.reference_state.dipole_moment
        elif level == 2:
            return self.mp2_dipole_moment
        else:
            raise NotImplementedError("Only dipole moments for level 0, 1 and 2"
                                      " are implemented.")

    @cached_member_function()
    def energy_correction(self, level=2):
        """Obtain the MP energy correction at a particular level"""
        if level > 3:
            raise NotImplementedError(f"MP({level}) energy correction "
                                      "not implemented.")
        if level < 2:
            return 0.0
        hf = self.reference_state
        is_cvs = self.has_core_occupied_space
        if level == 2 and not is_cvs:
            terms = [(1.0, hf.oovv, self.t2oo)]
        elif level == 2 and is_cvs:
            terms = [(1.0, hf.oovv, self.t2oo),
                     (2.0, hf.ocvv, self.t2oc),
                     (1.0, hf.ccvv, self.t2cc)]
        elif level == 3 and not is_cvs:
            terms = [(1.0, hf.oovv, self.td2(b.oovv))]
        elif level == 3 and is_cvs:
            raise NotImplementedError("CVS-MP3 energy correction not implemented.")
        return sum(
            -0.25 * pref * eri.dot(t2)
            for pref, eri, t2 in terms
        )

    def energy(self, level=2):
        """
        Obtain the total energy (SCF energy plus all corrections)
        at a particular level of perturbation theory.
        """
        if level == 0:
            # Sum of orbital energies ...
            raise NotImplementedError("Total MP(0) energy not implemented.")

        # Accumulator for all energy terms
        energies = [self.reference_state.energy_scf]
        for il in range(2, level + 1):
            energies.append(self.energy_correction(il))
        return sum(energies)

    def to_qcvars(self, properties=False, recurse=False, maxlevel=2):
        """
        Return a dictionary with property keys compatible to a Psi4 wavefunction
        or a QCEngine Atomicresults object.
        """
        qcvars = {}
        for level in range(2, maxlevel + 1):
            try:
                mpcorr = self.energy_correction(level)
                qcvars[f"MP{level} CORRELATION ENERGY"] = mpcorr
                qcvars[f"MP{level} TOTAL ENERGY"] = self.energy(level)
            except NotImplementedError:
                pass
            except ValueError:
                pass

        if properties:
            for level in range(2, maxlevel + 1):
                try:
                    qcvars["MP2 DIPOLE"] = self.dipole_moment(level)
                except NotImplementedError:
                    pass

        if recurse:
            qcvars.update(self.reference_state.to_qcvars(properties, recurse))
        return qcvars

    @property
    def mp2_density(self):
        return self.density(2)

    @cached_property
    def mp2_dipole_moment(self):
        refstate = self.reference_state
        dipole_integrals = refstate.operators.electric_dipole
        mp2corr = np.array([product_trace(comp, self.mp2_diffdm)
                            for comp in dipole_integrals])
        return refstate.dipole_moment + mp2corr

    @cached_member_function()
    def ssq(self, level=2):
        """
        Return <S^2> of the ground state.
        """
        if self.reference_state.restricted:
            raise NotImplementedError(
                "<S^2> is not implemented for restricted HF references."
            )
        ssq_1p_op = self.reference_state.operators.ssq_1p
        ssq_2p_op = self.reference_state.operators.ssq_2p
        # the trace of the second-order (and higher) correction to the RDM1
        # is zero -> no influence on top of HF density for ground state
        ssq_1p = product_trace(ssq_1p_op, self.density(0))
        ssq_2p = product_trace(ssq_2p_op, self.density_2p(level))
        return (ssq_1p + ssq_2p)


#
# Register cvs_p0 intermediate
#
@register_as_intermediate
def cvs_p0(hf, mp, intermediates):
    # NOTE: equal to mp2_diffdm if CVS applied for the density
    ret = OneParticleDensity(hf.mospaces, symmetry=OperatorSymmetry.HERMITIAN)
    ret.oo = -0.5 * einsum("ikab,jkab->ij", mp.t2oo, mp.t2oo)
    ret.ov = -0.5 * (+ einsum("ijbc,jabc->ia", mp.t2oo, hf.ovvv)
                     + einsum("jkib,jkab->ia", hf.ooov, mp.t2oo)) / mp.df(b.ov)
    ret.vv = 0.5 * einsum("ijac,ijbc->ab", mp.t2oo, mp.t2oo)
    return ret



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2018 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import os
import glob
import atexit
import shutil
import libadcc
import tempfile


class MemoryPool(libadcc.AdcMemory):
    def initialise(self, scratch_directory="/tmp", max_block_size=16,
                   allocator="standard"):
        """Initialise the adcc memory management.

        Parameters
        ----------
        scratch_directory : str, optional
            Directory for storing temporary pagefiles. This should be a fast
            storage location as tensor data will be mapped to asynchronously
            to this directory for the "libxc" allocator.

        max_block_size : int, optional
            The maximal size a tensor block may have along each axis.

        allocator : str, optional
            The allocator to be used. Valid values are "libxm" or "standard"
            (libstc++ allocator).
        """
        pagefile_directory = tempfile.mkdtemp(prefix="adcc_", dir=scratch_directory)
        super().initialise(pagefile_directory, max_block_size, allocator)
        atexit.register(MemoryPool.cleanup, self)

    def cleanup(self):
        if os.path.isdir(self.pagefile_directory):
            shutil.rmtree(self.pagefile_directory)

    @property
    def scratch_directory(self):
        return os.path.dirname(self.pagefile_directory)

    @property
    def page_files(self):
        """The list of all page files."""
        globs = ["pagefile.*", "xmpagefile"]
        globs = [os.path.join(self.pagefile_directory, pat) for pat in globs]
        return [c for pat in globs for c in glob.glob(pat)]

    @property
    def total_size_page_files(self):
        """The total size of all page files."""
        return sum(os.path.getsize(f) for f in self.page_files)


# The actual memory object to use
memory_pool = MemoryPool()



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2020 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------


class InputError(ValueError):
    """
    Exception thrown during the validation stage of the arguments passed to
    :py:`run_adc` to signal that an input is not valid.
    """
    pass


class InvalidReference(InputError):
    """
    Exception thrown if a passed SCF reference is invalid, e.g. because
    a feature like density-fitting has been applied, which is inconsistent
    with the current capabilities of adcc.
    """
    pass



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
import adcc
import psi4

mol = psi4.geometry("""
    O 0 0 0
    H 0 0 1.795239827225189
    H 1.693194615993441 0 -0.599043184453037
    symmetry c1
    units au
    """)

# set the number of cores equal to the auto-determined value from
# the adcc ThreadPool
psi4.set_num_threads(adcc.get_n_threads())
psi4.core.be_quiet()
psi4.set_options({'basis': "sto-3g",
                  'scf_type': 'pk',
                  'e_convergence': 1e-12,
                  'd_convergence': 1e-8})
scf_e, wfn = psi4.energy('SCF', return_wfn=True)

# Run an adc2 calculation:
state = adcc.adc2(wfn, n_singlets=5)

print(state.describe())



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2020 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import json
import itertools
import numpy as np

import requests


def write_commit_file(commit):
    with open("commit.rst", "w") as fp:
        fp.write("This summary only shows a few key results, which have been "
                 f"generated using commit **{commit}** from the "
                 "`adcc repository <https://code.adc-connect.org>`_. "
                 "The full results in interactive form are accessible on "
                 "https://adc-connect.github.io/adcc-bench.")


def fetch_json(commit, machine):
    baseurl = ("https://raw.githubusercontent.com/adc-connect/"
               "adcc-bench/results")
    resultsfile = (baseurl + "/results/" + machine + "/" + commit
                   + "-virtualenv-py3.6-numpy1.15-pybind11-""pyscf1.6.3"
                   "-scipy1.2.json")
    res = requests.get(resultsfile)
    assert res.ok
    return json.loads(res.text)


def write_details(data, testcase, reference, basis, n_ao=None):
    lines = []

    memkey = [k for k in data["results"].keys()
              if f"Full{testcase}" in k and "peakmem_" in k]
    timekey = [k for k in data["results"].keys()
               if f"Full{testcase}" in k and "time_" in k]
    if len(memkey) == 0:
        return
    assert len(memkey) == 1
    assert len(timekey) == 1
    memkey = memkey[0]
    timekey = timekey[0]

    mbasis, mmethod, mstates, mtol, mthreads = data["results"][memkey]["params"]
    tbasis, tmethod, tstates, ttol, tthreads = data["results"][timekey]["params"]
    tol = max(float(e) for e in set(mtol).intersection(ttol))
    states = max(int(e) for e in set(mstates).intersection(tstates))
    mproduct = list(itertools.product(*data["results"][memkey]["params"]))
    tproduct = list(itertools.product(*data["results"][timekey]["params"]))
    assert len(mthreads) == len(tthreads) == 1
    assert mthreads[0] == tthreads[0]
    threads = mthreads[0]

    kind = " (energies only)"
    if "oscillator_strength" in memkey:
        kind = " (energies and oscillator strength)"
    nfunc = ""
    if n_ao is not None:
        nfunc = f" ({n_ao} functions)"
    lines += [f"- **Basis set:** {basis}{nfunc}"]
    lines += [f"- **Reference:** {reference}"]
    lines += [f"- **Convergence tolerance:** {tol}"]
    lines += [f"- **Number of states:** {states} {kind}"]
    lines += [f"- **Threads:** {threads}"]
    lines += [""]

    lines += ["=========  =============  ===================="]
    lines += ["method          time (s)     peak memory (MiB)"]
    lines += ["=========  =============  ===================="]
    for method in sorted(set(mmethod).intersection(tmethod)):
        method = method[1:-1]
        for bas in [basis.upper(), basis.lower()]:
            try:
                midx = mproduct.index(("'" + bas + "'", "'" + method + "'",
                                       str(states), str(tol), str(threads)))
                tidx = tproduct.index(("'" + bas + "'", "'" + method + "'",
                                       str(states), str(tol), str(threads)))
            except ValueError:
                continue

        time = data["results"][timekey]["result"][tidx]
        memory = data["results"][memkey]["result"][midx]
        if time is None:
            time = np.nan
        if memory is not None:
            memory = memory / 1024**2
        else:
            memory = np.nan
        lines += [f"{method:<9s}  {time:>13.0f}  {memory:>20.0f}"]
    lines += ["=========  =============  ===================="]

    with open(testcase + ".rst", "w") as fp:
        fp.write("\n".join(lines))


def main():
    commit = "9c7bba83"
    write_commit_file(commit)
    json_clustern08 = fetch_json(commit, "mlv-clustern08")
    write_details(json_clustern08, "PhosphineCvs", "RHF", "6-311++G**", n_ao=51)
    write_details(json_clustern08, "MethylammoniumRadical", "UHF",
                  "cc-pVTZ", n_ao=116)
    write_details(json_clustern08, "ParaNitroAniline", "RHF", "cc-pVDZ", n_ao=170)
    write_details(json_clustern08, "WaterExpensive", "RHF", "cc-pVQZ", n_ao=115)
    write_details(json_clustern08, "Noradrenaline", "RHF", "6-311++G**", n_ao=341)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2021 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
from adcc import AmplitudeVector
from adcc.OneParticleDensity import OneParticleDensity
from adcc.NParticleOperator import OperatorSymmetry
from .matrix import AdcBlock


def block_ph_ph_0_pe(hf, mp, intermediates):
    """
    Constructs an :py:class:`AdcBlock` that describes the
    linear response coupling to the polarizable environment
    from PE via a CIS-like transition density as described
    in 10.1021/ct300763v, eq 63. Since the contribution
    depends on the input amplitude itself,
    a diagonal term cannot be formulated.
    """
    op = hf.operators

    def apply(ampl):
        tdm = OneParticleDensity(mp, symmetry=OperatorSymmetry.NOSYMMETRY)
        tdm.vo = ampl.ph.transpose()
        vpe = op.pe_induction_elec(tdm)
        return AmplitudeVector(ph=vpe.ov)
    return AdcBlock(apply, 0)


def block_ph_ph_0_pcm(hf, mp, intermediates):
    """
    Constructs an :py:class:`AdcBlock` that describes the
    linear response coupling to the polarizable environment
    from PCM via a CIS-like transition density as described
    in 10.1021/ct300763v, eq 63. Since the contribution
    depends on the input amplitude itself,
    a diagonal term cannot be formulated.
    """
    op = hf.operators

    def apply(ampl):
        tdm = OneParticleDensity(mp, symmetry=OperatorSymmetry.NOSYMMETRY)
        tdm.vo = ampl.ph.transpose()
        vpcm = op.pcm_potential_elec(tdm)
        return AmplitudeVector(ph=vpcm.ov)
    return AdcBlock(apply, 0)



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2020 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import numpy as np

from .OneParticleOperator import OneParticleOperator
from .StateView import StateView


class Excitation(StateView):
    def __init__(self, parent_state, index: int):
        """
        Construct an Excitation instance from an :class:`adcc.ExcitedStates`
        parent object.

        The class provides access to the properties of a single
        excited state, dynamically constructed inside ExcitedStates.excitations.

        Parameters
        ----------
        parent_state
            :class:`adcc.ExcitedStates` object from which the Excitation
            is derived
        index : int
            Index of the excited state the constructed :class:`adcc.Excitation`
            should refer to (0-based)
        """
        from .ExcitedStates import ElectronicTransition
        # NOTE: This should also work with S2S. But then index
        # would need to be relative to S2S.initial, i.e., 0 for S2S.initial + 1
        # which is kind of weird. But I think we can allow it.
        if not isinstance(parent_state, ElectronicTransition):
            raise TypeError("parent_state needs to be an ExcitedStates object. "
                            f"Got: {type(parent_state)}.")
        super().__init__(parent_state, index)
        self._parent_state: ElectronicTransition

    @property
    def transition_dm(self) -> OneParticleOperator:
        """The transition density matrix"""
        return self._parent_state._transition_dm(self.index)

    @property
    def transition_dm_ao(self) -> OneParticleOperator:
        """The transition density matrix in the AO basis"""
        return sum(self.transition_dm.to_ao_basis())

    @property
    def transition_dipole_moment(self) -> np.ndarray:
        """The transition dipole moment"""
        return self._parent_state._transition_dipole_moment(self.index)

    @property
    def transition_dipole_moment_velocity(self) -> np.ndarray:
        """The transition dipole moment in the velocity gauge"""
        return self._parent_state._transition_dipole_moment_velocity(self.index)

    def transition_magnetic_dipole_moment(self,
                                          gauge_origin="origin") -> np.ndarray:
        """The transition magnetic dipole moment"""
        return self._parent_state._transition_magnetic_dipole_moment(
            state_n=self.index, gauge_origin=gauge_origin
        )

    def transition_quadrupole_moment(self, gauge_origin="origin") -> np.ndarray:
        """The transition quadrupole moment"""
        return self._parent_state._transition_quadrupole_moment(
            state_n=self.index, gauge_origin=gauge_origin
        )

    def transition_quadrupole_moment_velocity(self,
                                              gauge_origin="origin") -> np.ndarray:
        """The transition quadrupole moment"""
        return self._parent_state._transition_quadrupole_moment_velocity(
            state_n=self.index, gauge_origin=gauge_origin
        )

    @property
    def oscillator_strength(self) -> np.float64:
        """The oscillator strength"""
        return self._parent_state._oscillator_strength(self.index)

    @property
    def oscillator_strength_velocity(self) -> np.float64:
        """The oscillator strengh in the velocity gauge"""
        return self._parent_state._oscillator_strength_velocity(self.index)

    @property
    def rotatory_strength(self) -> np.float64:
        """
        The rotatory strength (in velocity gauge).
        This property is gauge-origin invariant, thus, it is not possible to
        select a gauge origin.
        """
        return self._parent_state._rotatory_strength(self.index)

    def rotatory_strength_length(self, gauge_origin="origin") -> np.float64:
        """The rotatory strength in the length gauge"""
        return self._parent_state._rotatory_strength_length(
            state_n=self.index, gauge_origin=gauge_origin
        )

    @property
    def cross_section(self) -> np.float64:
        """The one-photon absorption cross section"""
        return self._parent_state._cross_section(self.index)



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
from pyscf import gto, scf
from matplotlib import pyplot as plt

import adcc

# pyscf-H2O Hartree-Fock calculation
mol = gto.M(
    atom='O 0 0 0;'
         'H 0 0 1.795239827225189;'
         'H 1.693194615993441 0 -0.599043184453037',
    basis='cc-pvtz',
    unit="Bohr"
)
scfres = scf.RHF(mol)
scfres.conv_tol = 1e-13
scfres.kernel()


plt.figure(figsize=(7, 5))
state = adcc.adc2(scfres, n_singlets=10)
state.plot_spectrum()

plt.savefig("plot_spectrum_water.png", bbox="tight")
plt.close()



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2018 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import numpy as np

from adcc.AdcMatrix import AdcMatrixlike
from adcc.AmplitudeVector import AmplitudeVector


class PreconditionerIdentity:
    """
    Preconditioner, which does absolutely nothing
    """
    def apply(self, invecs):
        return invecs

    def __matmul__(self, x):
        return x


class JacobiPreconditioner:
    """
    Jacobi-type preconditioner

    Represents the application of (D - σ I)^{-1}, where
    D is the diagonal of the adcmatrix.
    """
    def __init__(self, adcmatrix, shifts=0.0):
        if not isinstance(adcmatrix, AdcMatrixlike):
            raise TypeError("Only an AdcMatrixlike may be used with this "
                            "preconditioner for now.")

        self.diagonal = adcmatrix.diagonal()
        self.shifts = shifts

    def update_shifts(self, shifts):
        """
        Update the shift value or values applied to the diagonal.
        If this is a single value it will be applied to all
        vectors simultaneously. If it is multiple values,
        then each value will be applied only to one
        of the passed vectors.
        """
        self.shifts = shifts

    def apply(self, invecs):
        if isinstance(invecs, AmplitudeVector):
            if not isinstance(self.shifts, (float, np.number)):
                raise TypeError("Can only apply JacobiPreconditioner "
                                "to a single vector if shifts is "
                                "only a single number.")
            return invecs / (self.diagonal - self.shifts)
        elif isinstance(invecs, list):
            if len(self.shifts) != len(invecs):
                raise ValueError("Number of vectors passed does not agree "
                                 "with number of shifts stored inside "
                                 "precoditioner. Update using the "
                                 "'update_shifts' method.")

            return [v / (self.diagonal - self.shifts[i])
                    for i, v in enumerate(invecs)]
        else:
            raise TypeError("Input type not understood: " + str(type(invecs)))

    def __matmul__(self, invecs):
        return self.apply(invecs)

    # __matvec__



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2018 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
from collections import Counter
from dataclasses import dataclass
from typing import Any, Optional, Union, TypeVar
from enum import Enum


T = TypeVar("T", bound="Method")


class MethodLevel(Enum):
    # numeric levels
    ZERO = 0
    ONE = 1
    TWO = 2
    THREE = 3
    FOUR = 4
    FIVE = 5

    # special levels
    TWO_X = "2x"    # extended 2nd-order ADC: 2p2h-2p2h in 1st order
    # 1st-order ISR: in singles excitation space only (starting from 1-particle
    # operators, doubles are required for a consistent first-order description)
    ONE_S = "1s"
    # 2nd-order ISR: in doubles excitation space only (starting from 2-particle
    # operators, triples are required for a consistent second-order description)
    TWO_D = "2d"
    # 3rd-order ISR: in doubles excitation space only (starting from 1-particle
    # operators, triples are required for a consistent third-order description)
    THREE_D = "3d"

    def to_str(self) -> str:
        return str(self.value)

    def to_int(self) -> int:
        """
        Converts the level to an integer. This also resolves special
        levels. For instance, 'TWO_X' resolves as 2.
        """
        # numerical methods
        if isinstance(self.value, int):
            return self.value
        # return base int for special methods
        elif isinstance(self.value, str):
            return int(self.value[0])
        else:
            raise ValueError(f"Unknown value type {type(self.value)}.")


class AdcType(Enum):
    PP = "pp"

    def to_str(self) -> str:
        return self.value

    @classmethod
    def is_valid_adc_type(cls, adc_type: str) -> bool:
        try:
            cls(adc_type)
            return True
        except ValueError:
            return False


class GroundStateType(Enum):
    MP = "mp"

    def to_str(self) -> str:
        return self.value

    @classmethod
    def is_valid_gs_type(cls, gs_type: str) -> bool:
        try:
            cls(gs_type)
            return True
        except ValueError:
            return False


@dataclass(frozen=True)
class LevelSpec:
    max_level: Optional[int] = None
    special_levels: tuple[MethodLevel, ...] = tuple()

    def supports(self, level: MethodLevel) -> bool:
        if isinstance(level.value, int):
            return self.max_level is not None and level.value <= self.max_level
        return level in self.special_levels


@dataclass(frozen=True)
class LevelKey:
    adc_type: AdcType
    gs_type: GroundStateType
    cvs: bool


class Method:
    # this has to be set on the child classes
    _method_base_name: Optional[str] = None
    _supported_levels: dict[LevelKey, LevelSpec] = {}

    def __init__(self, method: str):
        """
        Constructs a Method validating and decomposing the given method string.
        Valid method strings are of the form
        [prefixes]-[adc_type]-[method_base_name][level]
        with the individual components separated by '-'.
        """
        assert self._method_base_name is not None

        # validate base method type, which has to be the last part of the
        # method string
        split = method.split("-")
        if not split[-1].startswith(self._method_base_name):
            raise ValueError(f"{split[-1]} is not a valid method type")

        # validate method level
        level = split[-1][len(self._method_base_name):]
        if level.isnumeric():
            self.level: MethodLevel = MethodLevel(int(level))
        else:
            self.level: MethodLevel = MethodLevel(level)

        split = split[:-1]
        # validate and set the adc_type, which has to be the last entry of split,
        # i.e., the method string has to end with e.g. ip-adc2
        self.adc_type: AdcType = AdcType.PP
        if split and AdcType.is_valid_adc_type(split[-1]):
            self.adc_type: AdcType = AdcType(split[-1])
            split = split[:-1]
        # Therefore, all remaining prefixes at this point can not be
        # a valid adc_type
        if any(AdcType.is_valid_adc_type(pref) for pref in split):
            raise ValueError(
                f"Invalid method string {method}. ADC type (e.g. 'pp') either "
                "provided twice or in wrong position. Valid method strings "
                f"have to end with e.g. 'pp-adc2'."
            )
        # validate prefixes
        valid_prefixes: tuple[str, ...] = ("cvs", "mp")
        if any(pref not in valid_prefixes for pref in split):
            raise ValueError(f"Invalid method prefix in {split}. Valid prefixes "
                             f"are {valid_prefixes}.")
        if any(count != 1 for count in Counter(split).values()):
            raise ValueError(f"Invalid method string {method}. Duplicate "
                             f"prefix detected in {split}.")
        # set and remove cvs
        self.is_core_valence_separated: bool = "cvs" in split
        if self.is_core_valence_separated:
            split.remove("cvs")
        # deal with the gs_type
        self.gs_type: GroundStateType = GroundStateType.MP
        for pref in split:
            if GroundStateType.is_valid_gs_type(pref):
                self.gs_type: GroundStateType = GroundStateType(pref)
                split.remove(pref)
                break
        # The remaining prefixes can not be a ground state type anymore
        if any(GroundStateType.is_valid_gs_type(pref) for pref in split):
            raise ValueError(f"Invalid method string {method}. Ground state "
                             "type (e.g. 'mp') provided twice.")
        # at this point all prefixes should have been handled and removed
        if split:
            raise ValueError(f"Invalid method prefixes detected: {split}."
                             f"Parsed from method string '{method}'.")
        # ensure that the level is valid for the given method
        # this also depends on the adc type, gs_type and possibly other prefixes
        self._validate_level(self.level)

    def _validate_level(self, level: MethodLevel) -> None:
        key = LevelKey(
            adc_type=self.adc_type, gs_type=self.gs_type,
            cvs=self.is_core_valence_separated
        )
        spec = self._supported_levels.get(key, None)
        if spec is not None and spec.supports(level):
            return
        raise NotImplementedError(f"{self.name} method is not implemented.")

    @property
    def name(self) -> str:
        """The name of the Method as string."""
        if self.prefixes:
            return f"{self.prefixes}-{self._base_method}"
        else:
            return self._base_method

    @property
    def _base_method(self) -> str:
        assert self._method_base_name is not None
        if self.adc_type is AdcType.PP:
            return f"{self._method_base_name}{self.level.to_str()}"
        else:
            return (
                f"{self.adc_type.to_str()}-"
                f"{self._method_base_name}{self.level.to_str()}"
            )

    @property
    def prefixes(self) -> str:
        """String containing all prefixes of the method separated by '-'."""
        ret = []
        if self.is_core_valence_separated:
            ret.append("cvs")
        if self.gs_type is not GroundStateType.MP:
            ret.append(self.gs_type.to_str())
        return "-".join(ret)

    @property
    def base_method(self: T) -> T:
        """
        The base (full) method, i.e. with all approximations such as
        CVS stripped off.
        """
        return self.__class__(self._base_method)

    def at_level(self: T, newlevel: Union[int, str]) -> T:
        """
        Return an equivalent method, where only the level is changed
        (e.g. calling this on a CVS method returns a CVS method)
        """
        assert self._method_base_name is not None
        if self.prefixes:
            return self.__class__(
                f"{self.prefixes}-{self._method_base_name}{newlevel}"
            )
        else:
            return self.__class__(
                f"{self._method_base_name}{newlevel}"
            )

    def as_method(self, method_cls: type[T]) -> T:
        """
        Return a equivalent Method with the method base name replaced
        by the provided name.
        """
        assert self._method_base_name is not None
        assert method_cls._method_base_name is not None
        return method_cls(
            self.name.replace(self._method_base_name, method_cls._method_base_name)
        )

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Method):
            return NotImplemented
        return self.name == other.name

    def __ne__(self, other: Any):
        if not isinstance(other, Method):
            return NotImplemented
        return self.name != other.name

    def __repr__(self) -> str:
        return f"Method(name={self.name})"


class AdcMethod(Method):
    _method_base_name = "adc"
    _supported_levels = {
        LevelKey(
            adc_type=AdcType.PP,
            gs_type=GroundStateType.MP,
            cvs=False
        ): LevelSpec(
            max_level=3,
            special_levels=(MethodLevel.TWO_X,)
        ),
        LevelKey(
            adc_type=AdcType.PP,
            gs_type=GroundStateType.MP,
            cvs=True
        ): LevelSpec(
            max_level=3,
            special_levels=(MethodLevel.TWO_X,)
        )
    }


class IsrMethod(Method):
    _method_base_name = "isr"
    _supported_levels = {
        LevelKey(
            adc_type=AdcType.PP,
            gs_type=GroundStateType.MP,
            cvs=False
        ): LevelSpec(
            max_level=2,
            special_levels=(MethodLevel.ONE_S, MethodLevel.TWO_D)
        ),
        LevelKey(
            adc_type=AdcType.PP,
            gs_type=GroundStateType.MP,
            cvs=True
        ): LevelSpec(
            max_level=2,
            special_levels=(MethodLevel.ONE_S, MethodLevel.TWO_D)
        )
    }



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
import adcc
import numpy as np

from scipy import constants
from matplotlib import pyplot as plt
from adcc.visualisation import ExcitationSpectrum

from pyscf import gto, scf
from pyscf.tools import cubegen

eV = constants.value("Hartree energy in eV")  # Hartree to eV

#
# This script shows a more low-level approach to property calculations
# showcasing how state and transition density matrices can be obtained
# and worked with. An example showing the recommended way to plot spectra
# can be found in pyscf_ccpvdz_adc2_spectrum.py
#


# Run SCF in pyscf
mol = gto.M(
    atom='O 0 0 0;'
         'H 0 0 1.795239827225189;'
         'H 1.693194615993441 0 -0.599043184453037',
    basis='cc-pvdz',
    unit="Bohr"
)
scfres = scf.RHF(mol)
scfres.conv_tol = 1e-12
scfres.conv_tol_grad = 1e-9
scfres.kernel()

print(adcc.banner())

# Run an adc2 calculation:
state = adcc.adc2(scfres, n_singlets=7, conv_tol=1e-8)

# Get dipole integrals from pyscf
dip_ao = mol.intor_symmetric('int1e_r', comp=3)

# compute nuclear dipole
charges = mol.atom_charges()
coords = mol.atom_coords()
dip_nucl = np.einsum('i,ix->x', charges, coords)

#
# Compute properties
#
exc_energies = []    # Excitation energies
osc_strengths = []    # Oscillator strength

print()
print("  st  ex.ene. (au)         f     transition dipole moment (au)"
      "        state dip (au)")
for i, exci in enumerate(state.excitation_energy):
    # Compute transition density matrix
    tdm_ao = state.transition_dm[i].to_ao_basis()
    ρ_tdm_tot = (tdm_ao[0] + tdm_ao[1]).to_ndarray()

    # Compute transition dipole moment
    tdip = np.einsum('xij,ij->x', dip_ao, ρ_tdm_tot)
    osc = 2. / 3. * np.linalg.norm(tdip)**2 * np.abs(exci)

    # Compute excited states density matrix and excited state dipole moment
    opdm_ao = state.state_dm[i].to_ao_basis()
    ρ_opdm_tot = (opdm_ao[0] + opdm_ao[1]).to_ndarray()
    sdip_el = np.einsum('xij,ij->x', dip_ao, ρ_opdm_tot)
    sdip = sdip_el - dip_nucl

    # Print findings
    fmt = "{0:2d}  {1:12.8g} {2:9.3g}   [{3:9.3g}, {4:9.3g}, {5:9.3g}]"
    fmt += "   [{6:9.3g}, {7:9.3g}, {8:9.3g}]"
    # fmt += "   [{9:9.3g}, {10:9.3g}, {11:9.3g}]"
    print(state.kind[0], fmt.format(i, exci, osc, *tdip, *sdip))

    # Build LUNTO and HONTO by SVD
    u, s, v = np.linalg.svd(ρ_tdm_tot)
    # LUNTOs
    cubegen.orbital(mol=mol, coeff=u.T[0],
                    outfile="nto_{}_LUNTO.cube".format(i))
    # HONTOs
    cubegen.orbital(mol=mol, coeff=v[0],
                    outfile="nto_{}_HONTO.cube".format(i))

    # Save oscillator strength and excitation energies
    osc_strengths.append(osc)
    exc_energies.append(state.excitation_energy[i])
exc_energies = np.array(exc_energies) * eV
osc_strengths = np.array(osc_strengths)

sp = ExcitationSpectrum(exc_energies, osc_strengths)
sp.xlabel = "Energy (eV)"
sp.plot(style="discrete", color="r")
sp_broad = sp.broaden_lines(shape="lorentzian", width=0.08)
sp_broad.plot(color="b", style="continuous")

plt.show()



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
import adcc
from pyscf import gto, scf

# Run SCF in pyscf
mol = gto.M(
    atom='H 0 0 0;'
         'F 0 0 2.5',
    basis='6-31G',
    unit="Bohr",
    spin=2  # =2S, ergo triplet
)
scfres = scf.UHF(mol)
scfres.conv_tol = 1e-14
scfres.conv_tol_grad = 1e-10
scfres.max_cycle = 500
scfres.kernel()

# Run solver and print results
states = adcc.adc2(scfres, n_spin_flip=5, conv_tol=1e-8)
print(states.describe())

# Plot the excitation spectrum
s2s = adcc.State2States(states, initial=0)
s2s.plot_spectrum()



#!/usr/bin/env python3
from pyscf import gto, mp, scf
from pyscf.geomopt import berny_solver

# Starting geometry
mol = gto.M(
    atom='S 0 0 0;'
         'H 0 0 1.795239827225189;'
         'H 1.693194615993441 0 -0.599043184453037',
    basis='cc-pvtz',
    verbose=3,
    unit="Bohr"
)

# HF optimisation
mf = scf.RHF(mol)
mol_hf_eq = berny_solver.optimize(mf)

# MP2 optimisation
mp2 = mp.MP2(scf.RHF(mol_hf_eq))
mol_mp2_eq = berny_solver.optimize(mp2)


print()
print("===========  Final MP2 geometry (bohr) ============")
print()
fmt = "{}  {:15.11g}  {:15.11g}  {:15.11g}"
coords = mol_mp2_eq.atom_coords()
n_atoms = len(coords)
for i in range(n_atoms):
    print(fmt.format(mol_mp2_eq.atom_symbol(i), *coords[i]))
